#!/usr/bin/env -S uv run

# /// script
# dependencies = ["nox>=2025.2.9"]
# ///

from __future__ import annotations

import argparse
import contextlib
import os
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator

import nox

nox.needs_version = ">=2025.2.9"
nox.options.default_venv_backend = "uv|virtualenv"


@nox.session(reuse_venv=True)
def lint(session: nox.Session) -> None:
    """
    Lint the codebase (except for clang-format/tidy).
    """
    session.install("pre-commit")
    session.run("pre-commit", "run", "-a", *session.posargs)


@nox.session
def tests(session: nox.Session) -> None:
    """
    Run the tests (requires a compiler).
    """
    tmpdir = session.create_tmp()
    session.install("cmake")
    session.install("-r", "tests/requirements.txt")
    session.run(
        "cmake",
        "-S.",
        f"-B{tmpdir}",
        "-DPYBIND11_WERROR=ON",
        "-DDOWNLOAD_CATCH=ON",
        "-DDOWNLOAD_EIGEN=ON",
        *session.posargs,
    )
    session.run("cmake", "--build", tmpdir)
    session.run("cmake", "--build", tmpdir, "--config=Release", "--target", "check")


@nox.session
def tests_packaging(session: nox.Session) -> None:
    """
    Run the packaging tests.
    """

    session.install("-r", "tests/requirements.txt", "pip")
    session.run("pytest", "tests/extra_python_package", *session.posargs)


@nox.session(reuse_venv=True, default=False)
def docs(session: nox.Session) -> None:
    """
    Build the docs. Pass --non-interactive to avoid serving.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-b", dest="builder", default="html", help="Build target (default: html)"
    )
    args, posargs = parser.parse_known_args(session.posargs)
    serve = args.builder == "html" and session.interactive

    extra_installs = ["sphinx-autobuild"] if serve else []
    session.install("-r", "docs/requirements.txt", *extra_installs)
    session.chdir("docs")

    shared_args = (
        "-n",  # nitpicky mode
        "-T",  # full tracebacks
        f"-b={args.builder}",
        ".",
        f"_build/{args.builder}",
        *posargs,
    )

    if serve:
        session.run(
            "sphinx-autobuild", "--open-browser", "--ignore=.build", *shared_args
        )
    else:
        session.run("sphinx-build", "--keep-going", *shared_args)


@nox.session(reuse_venv=True, default=False)
def make_changelog(session: nox.Session) -> None:
    """
    Inspect the closed issues and make entries for a changelog.
    """
    session.install_and_run_script("tools/make_changelog.py")


@nox.session(reuse_venv=True, default=False)
def build(session: nox.Session) -> None:
    """
    Build SDist and wheel.
    """

    session.install("build")
    session.log("Building normal files")
    session.run("python", "-m", "build", *session.posargs)


@contextlib.contextmanager
def preserve_file(filename: Path) -> Generator[str, None, None]:
    """
    Causes a file to be stored and preserved when the context manager exits.
    """
    old_stat = filename.stat()
    old_file = filename.read_text(encoding="utf-8")
    try:
        yield old_file
    finally:
        filename.write_text(old_file, encoding="utf-8")
        os.utime(filename, (old_stat.st_atime, old_stat.st_mtime))


@nox.session(reuse_venv=True)
def build_global(session: nox.Session) -> None:
    """
    Build global SDist and wheel.
    """

    installer = ["--installer=uv"] if session.venv_backend == "uv" else []
    session.install("build", "tomlkit")
    session.log("Building pybind11-global files")
    pyproject = Path("pyproject.toml")
    with preserve_file(pyproject):
        newer_txt = session.run("python", "tools/make_global.py", silent=True)
        assert isinstance(newer_txt, str)
        pyproject.write_text(newer_txt, encoding="utf-8")
        session.run(
            "python",
            "-m",
            "build",
            *installer,
            *session.posargs,
        )



"""Search for MRs and issues related to a list of commits."""

import argparse
import json
import sys
import subprocess
import re


def find_cherry_pick_source(commit_hash: str):
    """
    For a given commit hash, find the original commit it was cherry-picked from.

    Args:
        commit_hash: The commit hash to inspect.

    Returns:
        The full hash of the original commit if found, otherwise None.
    """
    try:
        # Use 'git show' to get the full commit message for the given hash.
        # The '-s' flag suppresses the diff output.
        # The '--format=%B' flag prints only the raw commit body/message.
        commit_message = subprocess.check_output(
            ["git", "show", "-s", "--format=%B", commit_hash.strip()],
            text=True,
            stderr=subprocess.PIPE,
        ).strip()

        # This regex looks for the specific line Git adds during a cherry-pick.
        # It captures the full 40-character SHA-1 hash.
        cherry_pick_pattern = re.compile(
            r"\(cherry picked from commit ([a-f0-9]{40})\)"
        )

        # Search the entire commit message for the pattern.
        match = cherry_pick_pattern.search(commit_message)

        if match:
            # If a match is found, return the captured group (the original commit hash).
            return match.group(1)
        else:
            return None

    except subprocess.CalledProcessError as e:
        # This error occurs if the git command fails, e.g., for an invalid hash.
        print(
            f"Error processing commit '{commit_hash.strip()}': {e.stderr.strip()}",
            file=sys.stderr,
        )
        return None
    except FileNotFoundError:
        # This error occurs if the 'git' command itself isn't found.
        print(
            "Error: 'git' command not found. Please ensure Git is installed and in your PATH.",
            file=sys.stderr,
        )
        sys.exit(1)


def main():
    """
    Main function to read commit hashes from stdin and process them.
    """
    parser = argparse.ArgumentParser(
        description="A script to download all MRs from GitLab matching specified criteria."
    )
    parser.add_argument(
        "--merge_requests_file",
        type=str,
        required=True,
        help="JSON file containing all the merge request information extracted via the GitLab API.",
    )

    # E.g. git log --pretty=%H 3e819d83bf52abda16bb53565f6801df40d071f1..3.4.1
    parser.add_argument(
        "--commits",
        required=True,
        help="List of commits, '-' for stdin.",
    )
    args = parser.parse_args()

    mrs = []
    with open(args.merge_requests_file, "r") as file:
        mrs = json.load(file)
        mrs_by_commit = {}

    if args.commits == "-":
        commit_hashes = sys.stdin.readlines()
    else:
        with open(args.commits, "r") as file:
            commit_hashes = file.readlines()

    # Arrange commits by SHA.
    for mr in mrs:
        for key in ["sha", "merge_commit_sha", "squash_commit_sha"]:
            sha = mr[key]
            if sha:
                mrs_by_commit[sha] = mr

    # Find the MRs and issues related to each commit.
    info = {}
    for sha in commit_hashes:
        sha = sha.strip()
        if not sha:
            continue

        # If a cherry-pick, extract the original hash.
        sha = find_cherry_pick_source(sha) or sha
        mr = mrs_by_commit.get(sha)

        commit_info = {}
        if mr:
            commit_info["merge_request"] = mr["iid"]
            commit_info["related_issues"] = [
                issue["iid"] for issue in mr["related_issues"]
            ]
            commit_info["closes_issues"] = [
                issue["iid"] for issue in mr["closes_issues"]
            ]

        info[sha] = commit_info

    print(json.dumps(info, indent=2))


if __name__ == "__main__":
    main()



from __future__ import annotations

import contextlib
import os
import re
import shutil
import subprocess
import sys
import tarfile
import zipfile
from pathlib import Path
from typing import Generator

# These tests must be run explicitly

DIR = Path(__file__).parent.resolve()
MAIN_DIR = DIR.parent.parent

FILENAME_VERSION = re.compile(r"[-_]((\d+\.\d+\.\d+)(?:[a-z]+\d*)?)(?:-|\.tar\.gz$)")

# Newer pytest has global path setting, but keeping old pytest for now
sys.path.append(str(MAIN_DIR / "tools"))

from make_global import get_global  # noqa: E402

HAS_UV = shutil.which("uv") is not None
UV_ARGS = ["--installer=uv"] if HAS_UV else []

PKGCONFIG = """\
prefix=${{pcfiledir}}/../../
includedir=${{prefix}}/include

Name: pybind11
Description: Seamless operability between C++11 and Python
Version: {VERSION}
Cflags: -I${{includedir}}
"""


main_headers = {
    "include/pybind11/attr.h",
    "include/pybind11/buffer_info.h",
    "include/pybind11/cast.h",
    "include/pybind11/chrono.h",
    "include/pybind11/common.h",
    "include/pybind11/complex.h",
    "include/pybind11/critical_section.h",
    "include/pybind11/eigen.h",
    "include/pybind11/embed.h",
    "include/pybind11/eval.h",
    "include/pybind11/functional.h",
    "include/pybind11/gil.h",
    "include/pybind11/gil_safe_call_once.h",
    "include/pybind11/gil_simple.h",
    "include/pybind11/iostream.h",
    "include/pybind11/native_enum.h",
    "include/pybind11/numpy.h",
    "include/pybind11/operators.h",
    "include/pybind11/options.h",
    "include/pybind11/pybind11.h",
    "include/pybind11/pytypes.h",
    "include/pybind11/subinterpreter.h",
    "include/pybind11/stl.h",
    "include/pybind11/stl_bind.h",
    "include/pybind11/trampoline_self_life_support.h",
    "include/pybind11/type_caster_pyobject_ptr.h",
    "include/pybind11/typing.h",
    "include/pybind11/warnings.h",
}

conduit_headers = {
    "include/pybind11/conduit/README.txt",
    "include/pybind11/conduit/pybind11_conduit_v1.h",
    "include/pybind11/conduit/pybind11_platform_abi_id.h",
    "include/pybind11/conduit/wrap_include_python_h.h",
}

detail_headers = {
    "include/pybind11/detail/class.h",
    "include/pybind11/detail/common.h",
    "include/pybind11/detail/cpp_conduit.h",
    "include/pybind11/detail/descr.h",
    "include/pybind11/detail/dynamic_raw_ptr_cast_if_possible.h",
    "include/pybind11/detail/function_record_pyobject.h",
    "include/pybind11/detail/init.h",
    "include/pybind11/detail/internals.h",
    "include/pybind11/detail/native_enum_data.h",
    "include/pybind11/detail/pybind11_namespace_macros.h",
    "include/pybind11/detail/struct_smart_holder.h",
    "include/pybind11/detail/type_caster_base.h",
    "include/pybind11/detail/typeid.h",
    "include/pybind11/detail/using_smart_holder.h",
    "include/pybind11/detail/value_and_holder.h",
    "include/pybind11/detail/exception_translation.h",
}

eigen_headers = {
    "include/pybind11/eigen/common.h",
    "include/pybind11/eigen/matrix.h",
    "include/pybind11/eigen/tensor.h",
}

stl_headers = {
    "include/pybind11/stl/filesystem.h",
}

cmake_files = {
    "share/cmake/pybind11/FindPythonLibsNew.cmake",
    "share/cmake/pybind11/pybind11Common.cmake",
    "share/cmake/pybind11/pybind11Config.cmake",
    "share/cmake/pybind11/pybind11ConfigVersion.cmake",
    "share/cmake/pybind11/pybind11GuessPythonExtSuffix.cmake",
    "share/cmake/pybind11/pybind11NewTools.cmake",
    "share/cmake/pybind11/pybind11Targets.cmake",
    "share/cmake/pybind11/pybind11Tools.cmake",
}

pkgconfig_files = {
    "share/pkgconfig/pybind11.pc",
}

py_files = {
    "__init__.py",
    "__main__.py",
    "_version.py",
    "commands.py",
    "py.typed",
    "setup_helpers.py",
    "share/__init__.py",
    "share/pkgconfig/__init__.py",
}

headers = main_headers | conduit_headers | detail_headers | eigen_headers | stl_headers
generated_files = cmake_files | pkgconfig_files
all_files = headers | generated_files | py_files

sdist_files = {
    "pyproject.toml",
    "LICENSE",
    "README.rst",
    "PKG-INFO",
    "SECURITY.md",
}


@contextlib.contextmanager
def preserve_file(filename: Path) -> Generator[str, None, None]:
    old_stat = filename.stat()
    old_file = filename.read_text(encoding="utf-8")
    try:
        yield old_file
    finally:
        filename.write_text(old_file, encoding="utf-8")
        os.utime(filename, (old_stat.st_atime, old_stat.st_mtime))


@contextlib.contextmanager
def build_global() -> Generator[None, None, None]:
    """
    Build global SDist and wheel.
    """

    pyproject = MAIN_DIR / "pyproject.toml"
    with preserve_file(pyproject):
        newer_txt = get_global()
        pyproject.write_text(newer_txt, encoding="utf-8")
        yield


def read_tz_file(tar: tarfile.TarFile, name: str) -> bytes:
    start = tar.getnames()[0].split("/")[0] + "/"
    inner_file = tar.extractfile(tar.getmember(f"{start}{name}"))
    assert inner_file
    with contextlib.closing(inner_file) as f:
        return f.read()


def normalize_line_endings(value: bytes) -> bytes:
    return value.replace(os.linesep.encode("utf-8"), b"\n")


def test_build_sdist(monkeypatch, tmpdir):
    monkeypatch.chdir(MAIN_DIR)

    subprocess.run(
        [sys.executable, "-m", "build", "--sdist", f"--outdir={tmpdir}", *UV_ARGS],
        check=True,
    )

    (sdist,) = tmpdir.visit("*.tar.gz")
    version = FILENAME_VERSION.search(sdist.basename).group(1)

    with tarfile.open(str(sdist), "r:gz") as tar:
        simpler = {n.split("/", 1)[-1] for n in tar.getnames()[1:]}
        (pkg_info_path,) = (n for n in simpler if n.endswith("PKG-INFO"))

        pyproject_toml = read_tz_file(tar, "pyproject.toml")
        pkg_info = read_tz_file(tar, pkg_info_path).decode("utf-8")

    files = headers | sdist_files
    assert files <= simpler

    assert b'name = "pybind11"' in pyproject_toml
    assert f"Version: {version}" in pkg_info
    assert "License-Expression: BSD-3-Clause" in pkg_info
    assert "License-File: LICENSE" in pkg_info
    assert "Provides-Extra: global" in pkg_info
    assert f'Requires-Dist: pybind11-global=={version}; extra == "global"' in pkg_info


def test_build_global_dist(monkeypatch, tmpdir):
    monkeypatch.chdir(MAIN_DIR)
    with build_global():
        subprocess.run(
            [
                sys.executable,
                "-m",
                "build",
                "--sdist",
                "--outdir",
                str(tmpdir),
                *UV_ARGS,
            ],
            check=True,
        )

    (sdist,) = tmpdir.visit("*.tar.gz")
    version = FILENAME_VERSION.search(sdist.basename).group(2)

    with tarfile.open(str(sdist), "r:gz") as tar:
        simpler = {n.split("/", 1)[-1] for n in tar.getnames()[1:]}
        (pkg_info_path,) = (n for n in simpler if n.endswith("PKG-INFO"))

        pyproject_toml = read_tz_file(tar, "pyproject.toml")
        pkg_info = read_tz_file(tar, pkg_info_path).decode("utf-8")

    files = headers | sdist_files
    assert files <= simpler

    assert b'name = "pybind11-global"' in pyproject_toml
    assert f"Version: {version}" in pkg_info
    assert "License-Expression: BSD-3-Clause" in pkg_info
    assert "License-File: LICENSE" in pkg_info
    assert "Provides-Extra: global" not in pkg_info
    assert 'Requires-Dist: pybind11-global; extra == "global"' not in pkg_info


def tests_build_wheel(monkeypatch, tmpdir):
    monkeypatch.chdir(MAIN_DIR)

    subprocess.run(
        [sys.executable, "-m", "build", "--wheel", "--outdir", str(tmpdir), *UV_ARGS],
        check=True,
    )

    (wheel,) = tmpdir.visit("*.whl")
    version, simple_version = FILENAME_VERSION.search(wheel.basename).groups()

    files = {f"pybind11/{n}" for n in all_files}
    files |= {
        "dist-info/licenses/LICENSE",
        "dist-info/METADATA",
        "dist-info/RECORD",
        "dist-info/WHEEL",
        "dist-info/entry_points.txt",
    }

    with zipfile.ZipFile(str(wheel)) as z:
        names = z.namelist()
        share = zipfile.Path(z, "pybind11/share")
        pkgconfig = (share / "pkgconfig/pybind11.pc").read_text(encoding="utf-8")
        cmakeconfig = (share / "cmake/pybind11/pybind11Config.cmake").read_text(
            encoding="utf-8"
        )
        (pkg_info_path,) = (n for n in names if n.endswith("METADATA"))
        pkg_info = zipfile.Path(z, pkg_info_path).read_text(encoding="utf-8")

    trimmed = {n for n in names if "dist-info" not in n}
    trimmed |= {f"dist-info/{n.split('/', 1)[-1]}" for n in names if "dist-info" in n}

    assert files == trimmed

    assert 'set(pybind11_INCLUDE_DIR "${PACKAGE_PREFIX_DIR}/include")' in cmakeconfig

    pkgconfig_expected = PKGCONFIG.format(VERSION=simple_version)
    assert pkgconfig_expected == pkgconfig

    assert f"Version: {version}" in pkg_info
    assert "License-Expression: BSD-3-Clause" in pkg_info
    assert "License-File: LICENSE" in pkg_info
    assert "Provides-Extra: global" in pkg_info
    assert f'Requires-Dist: pybind11-global=={version}; extra == "global"' in pkg_info


def tests_build_global_wheel(monkeypatch, tmpdir):
    monkeypatch.chdir(MAIN_DIR)
    with build_global():
        subprocess.run(
            [
                sys.executable,
                "-m",
                "build",
                "--wheel",
                "--outdir",
                str(tmpdir),
                *UV_ARGS,
            ],
            check=True,
        )

    (wheel,) = tmpdir.visit("*.whl")
    version, simple_version = FILENAME_VERSION.search(wheel.basename).groups()

    files = {f"data/data/{n}" for n in headers}
    files |= {f"data/headers/{n[8:]}" for n in headers}
    files |= {f"data/data/{n}" for n in generated_files}
    files |= {
        "dist-info/licenses/LICENSE",
        "dist-info/METADATA",
        "dist-info/WHEEL",
        "dist-info/RECORD",
    }

    with zipfile.ZipFile(str(wheel)) as z:
        names = z.namelist()
        beginning = names[0].split("/", 1)[0].rsplit(".", 1)[0]

        share = zipfile.Path(z, f"{beginning}.data/data/share")
        pkgconfig = (share / "pkgconfig/pybind11.pc").read_text(encoding="utf-8")
        cmakeconfig = (share / "cmake/pybind11/pybind11Config.cmake").read_text(
            encoding="utf-8"
        )

        (pkg_info_path,) = (n for n in names if n.endswith("METADATA"))
        pkg_info = zipfile.Path(z, pkg_info_path).read_text(encoding="utf-8")

    assert f"Version: {version}" in pkg_info
    assert "License-Expression: BSD-3-Clause" in pkg_info
    assert "License-File: LICENSE" in pkg_info
    assert "Provides-Extra: global" not in pkg_info
    assert 'Requires-Dist: pybind11-global; extra == "global"' not in pkg_info

    trimmed = {n[len(beginning) + 1 :] for n in names}

    assert files == trimmed

    assert 'set(pybind11_INCLUDE_DIR "${PACKAGE_PREFIX_DIR}/include")' in cmakeconfig

    pkgconfig_expected = PKGCONFIG.format(VERSION=simple_version)
    assert pkgconfig_expected == pkgconfig


def test_version_matches():
    header = MAIN_DIR / "include/pybind11/detail/common.h"
    text = header.read_text()

    # Extract the relevant macro values
    regex_prefix = r"#\s*define\s+PYBIND11_VERSION_"
    micro = re.search(rf"{regex_prefix}MICRO\s+(\d+)\b", text).group(1)
    release_level = re.search(rf"{regex_prefix}RELEASE_LEVEL\s+(\w+)\b", text).group(1)
    release_serial = re.search(
        rf"{regex_prefix}RELEASE_SERIAL\s+(\d+)\b",
        text,
    ).group(1)
    patch = re.search(rf"{regex_prefix}PATCH\s+([\w.-]+)\b", text).group(1)

    # Map release level macro to string
    level_map = {
        "PY_RELEASE_LEVEL_ALPHA": "a",
        "PY_RELEASE_LEVEL_BETA": "b",
        "PY_RELEASE_LEVEL_GAMMA": "rc",
        "PY_RELEASE_LEVEL_FINAL": "",
    }
    level_str = level_map[release_level]

    if release_level == "PY_RELEASE_LEVEL_FINAL":
        assert level_str == ""
        assert release_serial == "0"
        expected_patch = micro
    else:
        expected_patch = f"{micro}{level_str}{release_serial}"

    assert patch == expected_patch



from __future__ import annotations

import pickle
import re
import sys

import pytest

import env
from pybind11_tests import pickling as m


def all_pickle_protocols():
    assert pickle.HIGHEST_PROTOCOL >= 0
    return range(pickle.HIGHEST_PROTOCOL + 1)


@pytest.mark.parametrize("protocol", all_pickle_protocols())
def test_pickle_simple_callable(protocol):
    assert m.simple_callable() == 20220426
    serialized = pickle.dumps(m.simple_callable, protocol=protocol)
    assert b"pybind11_tests.pickling" in serialized
    assert b"simple_callable" in serialized
    deserialized = pickle.loads(serialized)
    assert deserialized() == 20220426
    assert deserialized is m.simple_callable

    # UNUSUAL: function record pickle roundtrip returns a module, not a function record object:
    if not env.PYPY:
        assert (
            pickle.loads(pickle.dumps(m.simple_callable.__self__, protocol=protocol))
            is m
        )
    # This is not expected to create issues because the only purpose of
    # `m.simple_callable.__self__` is to enable pickling: the only method it has is
    # `__reduce_ex__`. Direct access for any other purpose is not supported.
    # Note that `repr(m.simple_callable.__self__)` shows, e.g.:
    # `<pybind11_detail_function_record_v1__gcc_libstdcpp_cxxabi1018 object at 0x...>`
    # It is considered to be as much an implementation detail as the
    # `pybind11::detail::function_record` C++ type is.

    # @rainwoodman suggested that the unusual pickle roundtrip behavior could be
    # avoided by changing `reduce_ex_impl()` to produce, e.g.:
    # `"__import__('importlib').import_module('pybind11_tests.pickling').simple_callable.__self__"`
    # as the argument for the `eval()` function, and adding a getter to the
    # `function_record_PyTypeObject` that returns `self`. However, the additional code complexity
    # for this is deemed warranted only if the unusual pickle roundtrip behavior actually
    # creates issues.


@pytest.mark.parametrize("cls_name", ["Pickleable", "PickleableNew"])
def test_roundtrip(cls_name):
    cls = getattr(m, cls_name)
    p = cls("test_value")
    p.setExtra1(15)
    p.setExtra2(48)

    data = pickle.dumps(p, 2)  # Must use pickle protocol >= 2
    p2 = pickle.loads(data)
    assert p2.value() == p.value()
    assert p2.extra1() == p.extra1()
    assert p2.extra2() == p.extra2()


@pytest.mark.xfail("env.PYPY")
@pytest.mark.parametrize(
    "cls_name",
    [
        pytest.param(
            "PickleableWithDict",
            marks=pytest.mark.skipif(
                sys.version_info in ((3, 14, 0, "beta", 1), (3, 14, 0, "beta", 2)),
                reason="3.14.0b1/2 managed dict bug: https://github.com/python/cpython/issues/133912",
            ),
        ),
        "PickleableWithDictNew",
    ],
)
def test_roundtrip_with_dict(cls_name):
    cls = getattr(m, cls_name)
    p = cls("test_value")
    p.extra = 15
    p.dynamic = "Attribute"

    data = pickle.dumps(p, pickle.HIGHEST_PROTOCOL)
    p2 = pickle.loads(data)
    assert p2.value == p.value
    assert p2.extra == p.extra
    assert p2.dynamic == p.dynamic


def test_enum_pickle():
    from pybind11_tests import enums as e

    data = pickle.dumps(e.EOne, 2)
    assert e.EOne == pickle.loads(data)


#
# exercise_trampoline
#
class SimplePyDerived(m.SimpleBase):
    pass


def test_roundtrip_simple_py_derived():
    p = SimplePyDerived()
    p.num = 202
    p.stored_in_dict = 303
    data = pickle.dumps(p, pickle.HIGHEST_PROTOCOL)
    p2 = pickle.loads(data)
    assert isinstance(p2, SimplePyDerived)
    assert p2.num == 202
    assert p2.stored_in_dict == 303


def test_roundtrip_simple_cpp_derived():
    p = m.make_SimpleCppDerivedAsBase()
    assert m.check_dynamic_cast_SimpleCppDerived(p)
    p.num = 404
    if not env.PYPY:
        # To ensure that this unit test is not accidentally invalidated.
        with pytest.raises(AttributeError):
            # Mimics the `setstate` C++ implementation.
            setattr(p, "__dict__", {})  # noqa: B010
    data = pickle.dumps(p, pickle.HIGHEST_PROTOCOL)
    p2 = pickle.loads(data)
    assert isinstance(p2, m.SimpleBase)
    assert p2.num == 404
    # Issue #3062: pickleable base C++ classes can incur object slicing
    #              if derived typeid is not registered with pybind11
    assert not m.check_dynamic_cast_SimpleCppDerived(p2)


def test_new_style_pickle_getstate_pos_only():
    assert (
        re.match(
            r"^__getstate__\(self: [\w\.]+, /\)", m.PickleableNew.__getstate__.__doc__
        )
        is not None
    )
    if hasattr(m, "PickleableWithDictNew"):
        assert (
            re.match(
                r"^__getstate__\(self: [\w\.]+, /\)",
                m.PickleableWithDictNew.__getstate__.__doc__,
            )
            is not None
        )



# Adapted from:
# https://github.com/google/clif/blob/5718e4d0807fd3b6a8187dde140069120b81ecef/clif/testing/python/python_multiple_inheritance_test.py
from __future__ import annotations

from pybind11_tests import python_multiple_inheritance as m


class PC(m.CppBase):
    pass


class PPCC(PC, m.CppDrvd):
    pass


def test_PC():
    d = PC(11)
    assert d.get_base_value() == 11
    d.reset_base_value(13)
    assert d.get_base_value() == 13


def test_PPCC():
    d = PPCC(11)
    assert d.get_drvd_value() == 33
    d.reset_drvd_value(55)
    assert d.get_drvd_value() == 55

    assert d.get_base_value() == 11
    assert d.get_base_value_from_drvd() == 11
    d.reset_base_value(20)
    assert d.get_base_value() == 20
    assert d.get_base_value_from_drvd() == 20
    d.reset_base_value_from_drvd(30)
    assert d.get_base_value() == 30
    assert d.get_base_value_from_drvd() == 30



from __future__ import annotations

import pytest

from pybind11_tests import class_sh_mi_thunks as m


def test_ptrdiff_drvd_base0():
    ptrdiff = m.ptrdiff_drvd_base0()
    # A failure here does not (necessarily) mean that there is a bug, but that
    # test_class_sh_mi_thunks is not exercising what it is supposed to.
    # If this ever fails on some platforms: use pytest.skip()
    # If this ever fails on all platforms: don't know, seems extremely unlikely.
    assert ptrdiff != 0


@pytest.mark.parametrize(
    "vec_size_fn",
    [
        m.vec_size_base0_raw_ptr,
        m.vec_size_base0_shared_ptr,
    ],
)
@pytest.mark.parametrize(
    "get_fn",
    [
        m.get_drvd_as_base0_raw_ptr,
        m.get_drvd_as_base0_shared_ptr,
        m.get_drvd_as_base0_unique_ptr,
    ],
)
def test_get_vec_size_raw_shared(get_fn, vec_size_fn):
    obj = get_fn()
    assert vec_size_fn(obj) == 5


@pytest.mark.parametrize(
    "get_fn", [m.get_drvd_as_base0_raw_ptr, m.get_drvd_as_base0_unique_ptr]
)
def test_get_vec_size_unique(get_fn):
    obj = get_fn()
    assert m.vec_size_base0_unique_ptr(obj) == 5
    with pytest.raises(ValueError, match="Python instance was disowned"):
        m.vec_size_base0_unique_ptr(obj)


def test_get_shared_vec_size_unique():
    obj = m.get_drvd_as_base0_shared_ptr()
    with pytest.raises(ValueError) as exc_info:
        m.vec_size_base0_unique_ptr(obj)
    assert (
        str(exc_info.value) == "Cannot disown external shared_ptr (load_as_unique_ptr)."
    )



from __future__ import annotations

from pybind11_tests import class_sh_inheritance as m


def test_rtrn_mptr_drvd_pass_cptr_base():
    d = m.rtrn_mptr_drvd()
    i = m.pass_cptr_base(d)  # load_impl Case 2a
    assert i == 2 * 100 + 11


def test_rtrn_shmp_drvd_pass_shcp_base():
    d = m.rtrn_shmp_drvd()
    i = m.pass_shcp_base(d)  # load_impl Case 2a
    assert i == 2 * 100 + 21


def test_rtrn_mptr_drvd_up_cast_pass_cptr_drvd():
    b = m.rtrn_mptr_drvd_up_cast()
    # the base return is down-cast immediately.
    assert b.__class__.__name__ == "drvd"
    i = m.pass_cptr_drvd(b)
    assert i == 2 * 100 + 12


def test_rtrn_shmp_drvd_up_cast_pass_shcp_drvd():
    b = m.rtrn_shmp_drvd_up_cast()
    # the base return is down-cast immediately.
    assert b.__class__.__name__ == "drvd"
    i = m.pass_shcp_drvd(b)
    assert i == 2 * 100 + 22


def test_rtrn_mptr_drvd2_pass_cptr_bases():
    d = m.rtrn_mptr_drvd2()
    i1 = m.pass_cptr_base1(d)  # load_impl Case 2c
    assert i1 == 3 * 110 + 4 * 120 + 21
    i2 = m.pass_cptr_base2(d)
    assert i2 == 3 * 110 + 4 * 120 + 22


def test_rtrn_mptr_drvd2_up_casts_pass_cptr_drvd2():
    b1 = m.rtrn_mptr_drvd2_up_cast1()
    assert b1.__class__.__name__ == "drvd2"
    i1 = m.pass_cptr_drvd2(b1)
    assert i1 == 3 * 110 + 4 * 120 + 23
    b2 = m.rtrn_mptr_drvd2_up_cast2()
    assert b2.__class__.__name__ == "drvd2"
    i2 = m.pass_cptr_drvd2(b2)
    assert i2 == 3 * 110 + 4 * 120 + 23


def test_python_drvd2():
    class Drvd2(m.base1, m.base2):
        def __init__(self):
            m.base1.__init__(self)
            m.base2.__init__(self)

    d = Drvd2()
    i1 = m.pass_cptr_base1(d)  # load_impl Case 2b
    assert i1 == 110 + 21
    i2 = m.pass_cptr_base2(d)
    assert i2 == 120 + 22



from __future__ import annotations

import sys
import weakref

import pytest

import env
import pybind11_tests.class_sh_trampoline_shared_from_this as m


class PySft(m.Sft):
    pass


def test_release_and_shared_from_this():
    # Exercises the most direct path from building a shared_from_this-visible
    # shared_ptr to calling shared_from_this.
    obj = PySft("PySft")
    assert obj.history == "PySft"
    assert m.use_count(obj) == 1
    assert m.pass_shared_ptr(obj) == 2
    assert obj.history == "PySft_PassSharedPtr"
    assert m.use_count(obj) == 1
    assert m.pass_shared_ptr(obj) == 2
    assert obj.history == "PySft_PassSharedPtr_PassSharedPtr"
    assert m.use_count(obj) == 1


def test_release_and_shared_from_this_leak():
    obj = PySft("")
    while True:
        m.pass_shared_ptr(obj)
        assert not obj.history
        assert m.use_count(obj) == 1
        break  # Comment out for manual leak checking (use `top` command).


def test_release_and_stash():
    # Exercises correct functioning of guarded_delete weak_ptr.
    obj = PySft("PySft")
    stash1 = m.SftSharedPtrStash(1)
    stash1.Add(obj)
    exp_hist = "PySft_Stash1Add"
    assert obj.history == exp_hist
    assert m.use_count(obj) == 2
    assert stash1.history(0) == exp_hist
    assert stash1.use_count(0) == 1
    assert m.pass_shared_ptr(obj) == 3
    exp_hist += "_PassSharedPtr"
    assert obj.history == exp_hist
    assert m.use_count(obj) == 2
    assert stash1.history(0) == exp_hist
    assert stash1.use_count(0) == 1
    stash2 = m.SftSharedPtrStash(2)
    stash2.Add(obj)
    exp_hist += "_Stash2Add"
    assert obj.history == exp_hist
    assert m.use_count(obj) == 3
    assert stash2.history(0) == exp_hist
    assert stash2.use_count(0) == 2
    stash2.Add(obj)
    exp_hist += "_Stash2Add"
    assert obj.history == exp_hist
    assert m.use_count(obj) == 4
    assert stash1.history(0) == exp_hist
    assert stash1.use_count(0) == 3
    assert stash2.history(0) == exp_hist
    assert stash2.use_count(0) == 3
    assert stash2.history(1) == exp_hist
    assert stash2.use_count(1) == 3
    del obj
    assert stash2.history(0) == exp_hist
    assert stash2.use_count(0) == 3
    assert stash2.history(1) == exp_hist
    assert stash2.use_count(1) == 3
    stash2.Clear()
    assert stash1.history(0) == exp_hist
    assert stash1.use_count(0) == 1


def test_release_and_stash_leak():
    obj = PySft("")
    while True:
        stash1 = m.SftSharedPtrStash(1)
        stash1.Add(obj)
        assert not obj.history
        assert m.use_count(obj) == 2
        assert stash1.use_count(0) == 1
        stash1.Add(obj)
        assert not obj.history
        assert m.use_count(obj) == 3
        assert stash1.use_count(0) == 2
        assert stash1.use_count(1) == 2
        break  # Comment out for manual leak checking (use `top` command).


def test_release_and_stash_via_shared_from_this():
    # Exercises that the smart_holder vptr is invisible to the shared_from_this mechanism.
    obj = PySft("PySft")
    stash1 = m.SftSharedPtrStash(1)
    with pytest.raises(RuntimeError) as exc_info:
        stash1.AddSharedFromThis(obj)
    assert str(exc_info.value) == "bad_weak_ptr"
    stash1.Add(obj)
    assert obj.history == "PySft_Stash1Add"
    assert stash1.use_count(0) == 1
    stash1.AddSharedFromThis(obj)
    assert obj.history == "PySft_Stash1Add_Stash1AddSharedFromThis"
    assert stash1.use_count(0) == 2
    assert stash1.use_count(1) == 2


def test_release_and_stash_via_shared_from_this_leak():
    obj = PySft("")
    while True:
        stash1 = m.SftSharedPtrStash(1)
        with pytest.raises(RuntimeError) as exc_info:
            stash1.AddSharedFromThis(obj)
        assert str(exc_info.value) == "bad_weak_ptr"
        stash1.Add(obj)
        assert not obj.history
        assert stash1.use_count(0) == 1
        stash1.AddSharedFromThis(obj)
        assert not obj.history
        assert stash1.use_count(0) == 2
        assert stash1.use_count(1) == 2
        break  # Comment out for manual leak checking (use `top` command).


def test_pass_released_shared_ptr_as_unique_ptr():
    # Exercises that returning a unique_ptr fails while a shared_from_this
    # visible shared_ptr exists.
    obj = PySft("PySft")
    stash1 = m.SftSharedPtrStash(1)
    stash1.Add(obj)  # Releases shared_ptr to C++.
    assert m.pass_unique_ptr_cref(obj) == "PySft_Stash1Add"
    assert obj.history == "PySft_Stash1Add"
    with pytest.raises(ValueError) as exc_info:
        m.pass_unique_ptr_rref(obj)
    assert str(exc_info.value) == (
        "Python instance is currently owned by a std::shared_ptr."
    )
    assert obj.history == "PySft_Stash1Add"


@pytest.mark.parametrize(
    "make_f",
    [
        m.make_pure_cpp_sft_raw_ptr,
        m.make_pure_cpp_sft_unq_ptr,
        m.make_pure_cpp_sft_shd_ptr,
    ],
)
def test_pure_cpp_sft_raw_ptr(make_f):
    # Exercises void_cast_raw_ptr logic for different situations.
    obj = make_f("PureCppSft")
    assert m.pass_shared_ptr(obj) == 3
    assert obj.history == "PureCppSft_PassSharedPtr"
    obj = make_f("PureCppSft")
    stash1 = m.SftSharedPtrStash(1)
    stash1.AddSharedFromThis(obj)
    assert obj.history == "PureCppSft_Stash1AddSharedFromThis"


def test_multiple_registered_instances_for_same_pointee():
    obj0 = PySft("PySft")
    obj0.attachment_in_dict = "Obj0"
    assert m.pass_through_shd_ptr(obj0) is obj0
    while True:
        obj = m.Sft(obj0)
        assert obj is not obj0
        obj_pt = m.pass_through_shd_ptr(obj)
        # Unpredictable! Because registered_instances is as std::unordered_multimap.
        assert obj_pt is obj0 or obj_pt is obj
        # Multiple registered_instances for the same pointee can lead to unpredictable results:
        if obj_pt is obj0:
            assert obj_pt.attachment_in_dict == "Obj0"
        else:
            assert not hasattr(obj_pt, "attachment_in_dict")
        assert obj0.history == "PySft"
        break  # Comment out for manual leak checking (use `top` command).


def test_multiple_registered_instances_for_same_pointee_leak():
    obj0 = PySft("")
    while True:
        stash1 = m.SftSharedPtrStash(1)
        stash1.Add(m.Sft(obj0))
        assert stash1.use_count(0) == 1
        stash1.Add(m.Sft(obj0))
        assert stash1.use_count(0) == 1
        assert stash1.use_count(1) == 1
        assert not obj0.history
        break  # Comment out for manual leak checking (use `top` command).


def test_multiple_registered_instances_for_same_pointee_recursive():
    while True:
        obj0 = PySft("PySft")
        if not env.PYPY:
            obj0_wr = weakref.ref(obj0)
        obj = obj0
        # This loop creates a chain of instances linked by shared_ptrs.
        for _ in range(10):
            obj_next = m.Sft(obj)
            assert obj_next is not obj
            obj = obj_next
            del obj_next
            assert obj.history == "PySft"
        del obj0
        if not env.PYPY and not env.GRAALPY:
            assert obj0_wr() is not None
        del obj  # This releases the chain recursively.
        if not env.PYPY and not env.GRAALPY:
            assert obj0_wr() is None
        break  # Comment out for manual leak checking (use `top` command).


# As of 2021-07-10 the pybind11 GitHub Actions valgrind build uses Python 3.9.
WORKAROUND_ENABLING_ROLLBACK_OF_PR3068 = env.LINUX and sys.version_info == (3, 9)


def test_std_make_shared_factory():
    class PySftMakeShared(m.Sft):
        def __init__(self, history):
            super().__init__(history, 0)

    obj = PySftMakeShared("PySftMakeShared")
    assert obj.history == "PySftMakeShared"
    if WORKAROUND_ENABLING_ROLLBACK_OF_PR3068:
        try:
            m.pass_through_shd_ptr(obj)
        except RuntimeError as e:
            str_exc_info_value = str(e)
        else:
            str_exc_info_value = "RuntimeError NOT RAISED"
    else:
        with pytest.raises(RuntimeError) as exc_info:
            m.pass_through_shd_ptr(obj)
        str_exc_info_value = str(exc_info.value)
    assert (
        str_exc_info_value
        == "smart_holder_type_casters load_as_shared_ptr failure: not implemented:"
        " trampoline-self-life-support for external shared_ptr to type inheriting"
        " from std::enable_shared_from_this."
    )



from __future__ import annotations

from pybind11_tests import class_sh_shared_ptr_copy_move as m


def test_shptr_copy():
    txt = m.test_ShPtr_copy()[0].get_history()
    assert txt == "FooShPtr_copy"


def test_smhld_copy():
    txt = m.test_SmHld_copy()[0].get_history()
    assert txt == "FooSmHld_copy"


def test_shptr_move():
    txt = m.test_ShPtr_move()[0].get_history()
    assert txt == "FooShPtr_move"


def test_smhld_move():
    txt = m.test_SmHld_move()[0].get_history()
    assert txt == "FooSmHld_move"


def _check_property(foo_typ, prop_typ, policy):
    o = m.Outer()
    name = f"{foo_typ}_{prop_typ}_{policy}"
    history = f"Foo{foo_typ}_Outer"
    f = getattr(o, name)
    assert f.get_history() == history
    # and try again to check that o did not get changed
    f = getattr(o, name)
    assert f.get_history() == history


def test_properties():
    for prop_typ in ("readonly", "readwrite", "property_readonly"):
        for foo_typ in ("ShPtr", "SmHld"):
            for policy in ("default", "copy", "move"):
                _check_property(foo_typ, prop_typ, policy)



"""Downloads all MRs from GitLab matching specified criteria."""

import argparse
import datetime
import json
import os
import requests
import sys

EIGEN_PROJECT_ID = 15462818  # Taken from the gitlab project page.


def date(date_string: str):
    """Convert a date YY-MM-DD string to a datetime object."""
    try:
        return datetime.strptime(date_string, "%Y-%m-%d")
    except ValueError:
        msg = f"Not a valid date: '{date_string}'. Expected format is YYYY-MM-DD."
        raise argparse.ArgumentTypeError(msg)


def _get_api_query(
    gitlab_private_token: str, url: str, params: dict[str, str] | None = None
):
    next_page = "1"
    if not params:
        params = dict()
    params["per_page"] = "100"
    headers = {"PRIVATE-TOKEN": gitlab_private_token}
    out = []
    while next_page:
        params["page"] = next_page
        try:
            resp = requests.head(url=url, params=params, headers=headers)
            if resp.status_code != 200:
                print("Request failed: ", resp, file=sys.stderr)
                break

            next_next_page = resp.headers["x-next-page"]

            resp = requests.get(url=url, params=params, headers=headers)
            if resp.status_code != 200:
                # Try again.
                continue

            out.extend(resp.json())
            # Advance at the end, in case an exception occurs above so we can retry
            next_page = next_next_page
        except:
            # Keep same next_page
            continue
    return out


def get_merge_requests(
    gitlab_private_token: str,
    author_username: str | None = None,
    state: str | None = None,
    created_before: datetime.datetime | None = None,
    created_after: datetime.datetime | None = None,
    updated_after: datetime.datetime | None = None,
    updated_before: datetime.datetime | None = None,
    related_issues: bool = False,
    closes_issues: bool = False,
):
    """Return list of merge requests.

    Args:
      gitlab_token: GitLab API token.
      author_username: MR author username.
      state: MR state (merged, opened, closed, locked).
      created_after: datetime start of period.
      created_before: datetime end of period.
      updated_after: datetime start of period.
      updated_before: datetime end of period.

    Returns:
      List of merge requests.
    """
    url = (
        "https://gitlab.com/api/v4/projects/"
        + str(EIGEN_PROJECT_ID)
        + "/merge_requests"
    )
    params = dict()
    if author_username:
        params["author_username"] = author_username

    if state:
        params["state"] = state

    if created_before:
        params["created_before"] = created_before.isoformat()

    if created_after:
        params["created_after"] = created_after.isoformat()

    if updated_before:
        params["updated_before"] = updated_before.isoformat()

    if updated_after:
        params["updated_after"] = updated_after.isoformat()

    params["order_by"] = "created_at"
    params["sort"] = "asc"

    next_page = "1"
    params["per_page"] = "100"
    headers = {"PRIVATE-TOKEN": gitlab_private_token}

    mrs = _get_api_query(gitlab_private_token, url, params)

    if related_issues:
        for mr in mrs:
            mr["related_issues"] = _get_api_query(
                gitlab_private_token, f"{url}/{mr['iid']}/related_issues"
            )

    if closes_issues:
        for mr in mrs:
            mr["closes_issues"] = _get_api_query(
                gitlab_private_token, f"{url}/{mr['iid']}/closes_issues"
            )

    return mrs


def main(_):
    parser = argparse.ArgumentParser(
        description="A script to download all MRs from GitLab matching specified criteria."
    )
    parser.add_argument(
        "--gitlab_private_token",
        type=str,
        help="GitLab private API token.  Defaults to the GITLAB_PRIVATE_TOKEN environment variable if set.",
    )
    parser.add_argument("--author", type=str, help="The name of the author.")
    parser.add_argument(
        "--state",
        type=str,
        choices=["merged", "opened", "closed", "locked"],
        help="The state of the MR.",
    )
    parser.add_argument(
        "--created_before",
        type=date,
        help="The created-before date in YYYY-MM-DD format.",
    )
    parser.add_argument(
        "--created_after",
        type=date,
        help="The created-after date in YYYY-MM-DD format.",
    )
    parser.add_argument(
        "--updated_before",
        type=date,
        help="The updated-before date in YYYY-MM-DD format.",
    )
    parser.add_argument(
        "--updated_after",
        type=date,
        help="The updated-after date in YYYY-MM-DD format.",
    )
    parser.add_argument(
        "--related_issues", action="store_true", help="Query for related issues."
    )
    parser.add_argument(
        "--closes_issues",
        action="store_true",
        help="Query for issues closed by the MR.",
    )

    args = parser.parse_args()

    if not args.gitlab_private_token:
        args.gitlab_private_token = os.getenv("GITLAB_PRIVATE_TOKEN")
        if not args.gitlab_private_token:
            print("Could not determine GITLAB_PRIVATE_TOKEN.", file=sys.stderr)
            parser.print_usage()
            sys.exit(1)

    # Parse the arguments from the command line
    mrs = get_merge_requests(
        gitlab_private_token=args.gitlab_private_token,
        author_username=args.author,
        state=args.state,
        created_before=args.created_before,
        created_after=args.created_after,
        updated_before=args.updated_before,
        updated_after=args.updated_after,
        related_issues=args.related_issues,
        closes_issues=args.closes_issues,
    )

    mr_str = json.dumps(mrs, indent=2)
    print(mr_str)


if __name__ == "__main__":
    main(sys.argv)



# Copyright (c) 2025 The pybind Community.
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.

# This module tests the interaction of pybind11's shared_ptr and smart_holder
# mechanisms with trampoline object lifetime management and inheritance slicing.
#
# The following combinations are covered:
#
# - Holder type: std::shared_ptr (class_ holder) vs.
#                py::smart_holder
# - Conversion function: obj.cast<std::shared_ptr<T>>() vs.
#                        py::potentially_slicing_weak_ptr<T>(obj)
# - Python object type: C++ base class vs.
#                       Python-derived trampoline class
#
# The tests verify
#
# - that casting or passing Python objects into functions returns usable
#   std::shared_ptr<T> instances.
# - that inheritance slicing occurs as expected in controlled cases
#   (issue #1333).
# - that surprising weak_ptr behavior (issue #5623) can be reproduced when
#   smart_holder is used.
# - that the trampoline object remains alive in all situations
#   (no use-after-free) as long as the C++ shared_ptr exists.
#
# Where applicable, trampoline state is introspected to confirm whether the
# C++ object retains knowledge of the Python override or has fallen back to
# the base implementation.

from __future__ import annotations

import gc
import weakref

import pytest

import env
import pybind11_tests.potentially_slicing_weak_ptr as m


class PyDrvdSH(m.VirtBaseSH):
    def get_code(self):
        return 200


class PyDrvdSP(m.VirtBaseSP):
    def get_code(self):
        return 200


VIRT_BASE_TYPES = {
    "SH": {100: m.VirtBaseSH, 200: PyDrvdSH},
    "SP": {100: m.VirtBaseSP, 200: PyDrvdSP},
}

RTRN_FUNCS = {
    "SH": {
        "oc": m.SH_rtrn_obj_cast_shared_ptr,
        "ps": m.SH_rtrn_potentially_slicing_shared_ptr,
    },
    "SP": {
        "oc": m.SP_rtrn_obj_cast_shared_ptr,
        "ps": m.SP_rtrn_potentially_slicing_shared_ptr,
    },
}

SP_OWNER_TYPES = {
    "SH": m.SH_SpOwner,
    "SP": m.SP_SpOwner,
}

WP_OWNER_TYPES = {
    "SH": m.SH_WpOwner,
    "SP": m.SP_WpOwner,
}

GC_IS_RELIABLE = not (env.PYPY or env.GRAALPY)


@pytest.mark.parametrize("expected_code", [100, 200])
@pytest.mark.parametrize("rtrn_kind", ["oc", "ps"])
@pytest.mark.parametrize("holder_kind", ["SH", "SP"])
def test_rtrn_obj_cast_shared_ptr(holder_kind, rtrn_kind, expected_code):
    obj = VIRT_BASE_TYPES[holder_kind][expected_code]()
    ptr = RTRN_FUNCS[holder_kind][rtrn_kind](obj)
    assert ptr.get_code() == expected_code
    objref = weakref.ref(obj)
    del obj
    gc.collect()
    assert ptr.get_code() == expected_code  # the ptr Python object keeps obj alive
    assert objref() is not None
    del ptr
    gc.collect()
    if GC_IS_RELIABLE:
        assert objref() is None


@pytest.mark.parametrize("expected_code", [100, 200])
@pytest.mark.parametrize("holder_kind", ["SH", "SP"])
def test_with_sp_owner(holder_kind, expected_code):
    spo = SP_OWNER_TYPES[holder_kind]()
    assert spo.get_code() == -888
    assert spo.get_trampoline_state() == "sp nullptr"

    obj = VIRT_BASE_TYPES[holder_kind][expected_code]()
    assert obj.get_code() == expected_code

    spo.set_sp(obj)
    assert spo.get_code() == expected_code
    expected_trampoline_state = (
        "dynamic_cast failed" if expected_code == 100 else "trampoline alive"
    )
    assert spo.get_trampoline_state() == expected_trampoline_state

    del obj
    gc.collect()
    if holder_kind == "SH":
        assert spo.get_code() == expected_code
    elif GC_IS_RELIABLE:
        assert (
            spo.get_code() == 100
        )  # see issue #1333 (inheritance slicing) and PR #5624
    assert spo.get_trampoline_state() == expected_trampoline_state


@pytest.mark.parametrize("expected_code", [100, 200])
@pytest.mark.parametrize("set_meth", ["set_wp", "set_wp_potentially_slicing"])
@pytest.mark.parametrize("holder_kind", ["SH", "SP"])
def test_with_wp_owner(holder_kind, set_meth, expected_code):
    wpo = WP_OWNER_TYPES[holder_kind]()
    assert wpo.get_code() == -999
    assert wpo.get_trampoline_state() == "sp nullptr"

    obj = VIRT_BASE_TYPES[holder_kind][expected_code]()
    assert obj.get_code() == expected_code

    getattr(wpo, set_meth)(obj)
    if (
        holder_kind == "SP"
        or expected_code == 100
        or set_meth == "set_wp_potentially_slicing"
    ):
        assert wpo.get_code() == expected_code
    else:
        assert wpo.get_code() == -999  # see issue #5623 (weak_ptr expired) and PR #5624
    if expected_code == 100:
        expected_trampoline_state = "dynamic_cast failed"
    elif holder_kind == "SH" and set_meth == "set_wp":
        expected_trampoline_state = "sp nullptr"
    else:
        expected_trampoline_state = "trampoline alive"
    assert wpo.get_trampoline_state() == expected_trampoline_state

    del obj
    gc.collect()
    if GC_IS_RELIABLE:
        assert wpo.get_code() == -999


def test_potentially_slicing_weak_ptr_not_convertible_error():
    with pytest.raises(Exception) as excinfo:
        m.SH_rtrn_potentially_slicing_shared_ptr("")
    assert str(excinfo.value) == (
        '"str" object is not convertible to std::weak_ptr<T>'
        " (with T = pybind11_tests::potentially_slicing_weak_ptr::VirtBase<0>)"
    )
    with pytest.raises(Exception) as excinfo:
        m.SP_rtrn_potentially_slicing_shared_ptr([])
    assert str(excinfo.value) == (
        '"list" object is not convertible to std::weak_ptr<T>'
        " (with T = pybind11_tests::potentially_slicing_weak_ptr::VirtBase<1>)"
    )



from __future__ import annotations

from pybind11_tests import class_sh_trampoline_basic as m


class PyDrvd(m.Abase):
    def __init__(self, val):
        super().__init__(val)

    def Add(self, other_val):
        return self.Get() * 100 + other_val


def test_drvd_add():
    drvd = PyDrvd(74)
    assert drvd.Add(38) == (74 * 10 + 3) * 100 + 38


def test_drvd_add_in_cpp_raw_ptr():
    drvd = PyDrvd(52)
    assert m.AddInCppRawPtr(drvd, 27) == ((52 * 10 + 3) * 100 + 27) * 10 + 7


def test_drvd_add_in_cpp_shared_ptr():
    while True:
        drvd = PyDrvd(36)
        assert m.AddInCppSharedPtr(drvd, 56) == ((36 * 10 + 3) * 100 + 56) * 100 + 11
        return  # Comment out for manual leak checking (use `top` command).


def test_drvd_add_in_cpp_unique_ptr():
    while True:
        drvd = PyDrvd(25)
        assert m.AddInCppUniquePtr(drvd, 83) == ((25 * 10 + 3) * 100 + 83) * 100 + 13
        return  # Comment out for manual leak checking (use `top` command).



from __future__ import annotations

import pytest

from pybind11_tests import scoped_critical_section as m


def test_nullptr_combinations():
    m.test_one_nullptr()
    m.test_two_nullptrs()
    m.test_first_nullptr()
    m.test_second_nullptr()


@pytest.mark.skipif(not m.has_barrier, reason="no <barrier>")
def test_scoped_critical_section() -> None:
    for _ in range(64):
        m.test_scoped_critical_section()


@pytest.mark.skipif(not m.has_barrier, reason="no <barrier>")
def test_scoped_critical_section2() -> None:
    for _ in range(64):
        m.test_scoped_critical_section2()


@pytest.mark.skipif(not m.has_barrier, reason="no <barrier>")
def test_scoped_critical_section2_same_object_no_deadlock() -> None:
    for _ in range(64):
        m.test_scoped_critical_section2_same_object_no_deadlock()



from __future__ import annotations

import pytest

import env  # noqa: F401
from pybind11_tests import ConstructorStats, UserType
from pybind11_tests import stl as m


def test_vector(doc):
    """std::vector <-> list"""
    lst = m.cast_vector()
    assert lst == [1]
    lst.append(2)
    assert m.load_vector(lst)
    assert m.load_vector(tuple(lst))

    assert m.cast_bool_vector() == [True, False]
    assert m.load_bool_vector([True, False])
    assert m.load_bool_vector((True, False))

    assert doc(m.cast_vector) == "cast_vector() -> list[int]"
    assert (
        doc(m.load_vector)
        == "load_vector(arg0: collections.abc.Sequence[typing.SupportsInt]) -> bool"
    )

    # Test regression caused by 936: pointers to stl containers weren't castable
    assert m.cast_ptr_vector() == ["lvalue", "lvalue"]


def test_deque():
    """std::deque <-> list"""
    lst = m.cast_deque()
    assert lst == [1]
    lst.append(2)
    assert m.load_deque(lst)
    assert m.load_deque(tuple(lst))


def test_array(doc):
    """std::array <-> list"""
    lst = m.cast_array()
    assert lst == [1, 2]
    assert m.load_array(lst)
    assert m.load_array(tuple(lst))

    assert (
        doc(m.cast_array)
        == 'cast_array() -> typing.Annotated[list[int], "FixedSize(2)"]'
    )
    assert (
        doc(m.load_array)
        == 'load_array(arg0: typing.Annotated[collections.abc.Sequence[typing.SupportsInt], "FixedSize(2)"]) -> bool'
    )


def test_array_no_default_ctor():
    lst = m.NoDefaultCtorArray(3)
    assert [e.val for e in lst.arr] == [13, 23]
    lst.arr = m.NoDefaultCtorArray(4).arr
    assert [e.val for e in lst.arr] == [14, 24]


def test_valarray(doc):
    """std::valarray <-> list"""
    lst = m.cast_valarray()
    assert lst == [1, 4, 9]
    assert m.load_valarray(lst)
    assert m.load_valarray(tuple(lst))

    assert doc(m.cast_valarray) == "cast_valarray() -> list[int]"
    assert (
        doc(m.load_valarray)
        == "load_valarray(arg0: collections.abc.Sequence[typing.SupportsInt]) -> bool"
    )


def test_map(doc):
    """std::map <-> dict"""
    d = m.cast_map()
    assert d == {"key": "value"}
    assert "key" in d
    d["key2"] = "value2"
    assert "key2" in d
    assert m.load_map(d)

    assert doc(m.cast_map) == "cast_map() -> dict[str, str]"
    assert (
        doc(m.load_map) == "load_map(arg0: collections.abc.Mapping[str, str]) -> bool"
    )


def test_set(doc):
    """std::set <-> set"""
    s = m.cast_set()
    assert s == {"key1", "key2"}
    s.add("key3")
    assert m.load_set(s)
    assert m.load_set(frozenset(s))

    assert doc(m.cast_set) == "cast_set() -> set[str]"
    assert doc(m.load_set) == "load_set(arg0: collections.abc.Set[str]) -> bool"


def test_recursive_casting():
    """Tests that stl casters preserve lvalue/rvalue context for container values"""
    assert m.cast_rv_vector() == ["rvalue", "rvalue"]
    assert m.cast_lv_vector() == ["lvalue", "lvalue"]
    assert m.cast_rv_array() == ["rvalue", "rvalue", "rvalue"]
    assert m.cast_lv_array() == ["lvalue", "lvalue"]
    assert m.cast_rv_map() == {"a": "rvalue"}
    assert m.cast_lv_map() == {"a": "lvalue", "b": "lvalue"}
    assert m.cast_rv_nested() == [[[{"b": "rvalue", "c": "rvalue"}], [{"a": "rvalue"}]]]
    assert m.cast_lv_nested() == {
        "a": [[["lvalue", "lvalue"]], [["lvalue", "lvalue"]]],
        "b": [[["lvalue", "lvalue"], ["lvalue", "lvalue"]]],
    }

    # Issue #853 test case:
    z = m.cast_unique_ptr_vector()
    assert z[0].value == 7
    assert z[1].value == 42


def test_move_out_container():
    """Properties use the `reference_internal` policy by default. If the underlying function
    returns an rvalue, the policy is automatically changed to `move` to avoid referencing
    a temporary. In case the return value is a container of user-defined types, the policy
    also needs to be applied to the elements, not just the container."""
    c = m.MoveOutContainer()
    moved_out_list = c.move_list
    assert [x.value for x in moved_out_list] == [0, 1, 2]


@pytest.mark.skipif(not hasattr(m, "has_optional"), reason="no <optional>")
def test_optional():
    assert m.double_or_zero(None) == 0
    assert m.double_or_zero(42) == 84
    pytest.raises(TypeError, m.double_or_zero, "foo")

    assert m.half_or_none(0) is None
    assert m.half_or_none(42) == 21
    pytest.raises(TypeError, m.half_or_none, "foo")

    assert m.test_nullopt() == 42
    assert m.test_nullopt(None) == 42
    assert m.test_nullopt(42) == 42
    assert m.test_nullopt(43) == 43

    assert m.test_no_assign() == 42
    assert m.test_no_assign(None) == 42
    assert m.test_no_assign(m.NoAssign(43)) == 43
    pytest.raises(TypeError, m.test_no_assign, 43)

    assert m.nodefer_none_optional(None)

    holder = m.OptionalHolder()
    mvalue = holder.member
    assert mvalue.initialized
    assert holder.member_initialized()

    props = m.OptionalProperties()
    assert int(props.access_by_ref) == 42
    assert int(props.access_by_copy) == 42


@pytest.mark.skipif(
    not hasattr(m, "has_exp_optional"), reason="no <experimental/optional>"
)
def test_exp_optional():
    assert m.double_or_zero_exp(None) == 0
    assert m.double_or_zero_exp(42) == 84
    pytest.raises(TypeError, m.double_or_zero_exp, "foo")

    assert m.half_or_none_exp(0) is None
    assert m.half_or_none_exp(42) == 21
    pytest.raises(TypeError, m.half_or_none_exp, "foo")

    assert m.test_nullopt_exp() == 42
    assert m.test_nullopt_exp(None) == 42
    assert m.test_nullopt_exp(42) == 42
    assert m.test_nullopt_exp(43) == 43

    assert m.test_no_assign_exp() == 42
    assert m.test_no_assign_exp(None) == 42
    assert m.test_no_assign_exp(m.NoAssign(43)) == 43
    pytest.raises(TypeError, m.test_no_assign_exp, 43)

    holder = m.OptionalExpHolder()
    mvalue = holder.member
    assert mvalue.initialized
    assert holder.member_initialized()

    props = m.OptionalExpProperties()
    assert int(props.access_by_ref) == 42
    assert int(props.access_by_copy) == 42


@pytest.mark.skipif(not hasattr(m, "has_boost_optional"), reason="no <boost/optional>")
def test_boost_optional():
    assert m.double_or_zero_boost(None) == 0
    assert m.double_or_zero_boost(42) == 84
    pytest.raises(TypeError, m.double_or_zero_boost, "foo")

    assert m.half_or_none_boost(0) is None
    assert m.half_or_none_boost(42) == 21
    pytest.raises(TypeError, m.half_or_none_boost, "foo")

    assert m.test_nullopt_boost() == 42
    assert m.test_nullopt_boost(None) == 42
    assert m.test_nullopt_boost(42) == 42
    assert m.test_nullopt_boost(43) == 43

    assert m.test_no_assign_boost() == 42
    assert m.test_no_assign_boost(None) == 42
    assert m.test_no_assign_boost(m.NoAssign(43)) == 43
    pytest.raises(TypeError, m.test_no_assign_boost, 43)

    holder = m.OptionalBoostHolder()
    mvalue = holder.member
    assert mvalue.initialized
    assert holder.member_initialized()

    props = m.OptionalBoostProperties()
    assert int(props.access_by_ref) == 42
    assert int(props.access_by_copy) == 42


def test_reference_sensitive_optional(doc):
    assert m.double_or_zero_refsensitive(None) == 0
    assert m.double_or_zero_refsensitive(42) == 84
    pytest.raises(TypeError, m.double_or_zero_refsensitive, "foo")

    assert (
        doc(m.double_or_zero_refsensitive)
        == "double_or_zero_refsensitive(arg0: typing.SupportsInt | None) -> int"
    )

    assert m.half_or_none_refsensitive(0) is None
    assert m.half_or_none_refsensitive(42) == 21
    pytest.raises(TypeError, m.half_or_none_refsensitive, "foo")

    assert m.test_nullopt_refsensitive() == 42
    assert m.test_nullopt_refsensitive(None) == 42
    assert m.test_nullopt_refsensitive(42) == 42
    assert m.test_nullopt_refsensitive(43) == 43

    assert m.test_no_assign_refsensitive() == 42
    assert m.test_no_assign_refsensitive(None) == 42
    assert m.test_no_assign_refsensitive(m.NoAssign(43)) == 43
    pytest.raises(TypeError, m.test_no_assign_refsensitive, 43)

    holder = m.OptionalRefSensitiveHolder()
    mvalue = holder.member
    assert mvalue.initialized
    assert holder.member_initialized()

    props = m.OptionalRefSensitiveProperties()
    assert int(props.access_by_ref) == 42
    assert int(props.access_by_copy) == 42


@pytest.mark.skipif(not hasattr(m, "has_filesystem"), reason="no <filesystem>")
def test_fs_path():
    from pathlib import Path

    class PseudoStrPath:
        def __fspath__(self):
            return "foo/bar"

    class PseudoBytesPath:
        def __fspath__(self):
            return b"foo/bar"

    # Single argument
    assert m.parent_path(Path("foo/bar")) == Path("foo")
    assert m.parent_path("foo/bar") == Path("foo")
    assert m.parent_path(b"foo/bar") == Path("foo")
    assert m.parent_path(PseudoStrPath()) == Path("foo")
    assert m.parent_path(PseudoBytesPath()) == Path("foo")
    # std::vector
    assert m.parent_paths(["foo/bar", "foo/baz"]) == [Path("foo"), Path("foo")]
    # py::typing::List
    assert m.parent_paths_list(["foo/bar", "foo/baz"]) == [Path("foo"), Path("foo")]
    # Nested py::typing::List
    assert m.parent_paths_nested_list([["foo/bar"], ["foo/baz", "foo/buzz"]]) == [
        [Path("foo")],
        [Path("foo"), Path("foo")],
    ]
    # py::typing::Tuple
    assert m.parent_paths_tuple(("foo/bar", "foo/baz")) == (Path("foo"), Path("foo"))
    # py::typing::Dict
    assert m.parent_paths_dict(
        {
            "key1": Path("foo/bar"),
            "key2": "foo/baz",
            "key3": b"foo/buzz",
        }
    ) == {
        "key1": Path("foo"),
        "key2": Path("foo"),
        "key3": Path("foo"),
    }


@pytest.mark.skipif(not hasattr(m, "has_filesystem"), reason="no <filesystem>")
def test_path_typing(doc):
    # Single argument
    assert (
        doc(m.parent_path)
        == "parent_path(arg0: os.PathLike | str | bytes) -> pathlib.Path"
    )
    # std::vector
    assert (
        doc(m.parent_paths)
        == "parent_paths(arg0: collections.abc.Sequence[os.PathLike | str | bytes]) -> list[pathlib.Path]"
    )
    # py::typing::List
    assert (
        doc(m.parent_paths_list)
        == "parent_paths_list(arg0: list[os.PathLike | str | bytes]) -> list[pathlib.Path]"
    )
    # Nested py::typing::List
    assert (
        doc(m.parent_paths_nested_list)
        == "parent_paths_nested_list(arg0: list[list[os.PathLike | str | bytes]]) -> list[list[pathlib.Path]]"
    )
    # py::typing::Tuple
    assert (
        doc(m.parent_paths_tuple)
        == "parent_paths_tuple(arg0: tuple[os.PathLike | str | bytes, os.PathLike | str | bytes]) -> tuple[pathlib.Path, pathlib.Path]"
    )
    # py::typing::Dict
    assert (
        doc(m.parent_paths_dict)
        == "parent_paths_dict(arg0: dict[str, os.PathLike | str | bytes]) -> dict[str, pathlib.Path]"
    )


@pytest.mark.skipif(not hasattr(m, "load_variant"), reason="no <variant>")
def test_variant(doc):
    assert m.load_variant(1) == "int"
    assert m.load_variant("1") == "std::string"
    assert m.load_variant(1.0) == "double"
    assert m.load_variant(None) == "std::nullptr_t"

    assert m.load_variant_2pass(1) == "int"
    assert m.load_variant_2pass(1.0) == "double"

    assert m.cast_variant() == (5, "Hello")

    assert (
        doc(m.load_variant)
        == "load_variant(arg0: typing.SupportsInt | str | typing.SupportsFloat | None) -> str"
    )


@pytest.mark.skipif(
    not hasattr(m, "load_monostate_variant"), reason="no std::monostate"
)
def test_variant_monostate(doc):
    assert m.load_monostate_variant(None) == "std::monostate"
    assert m.load_monostate_variant(1) == "int"
    assert m.load_monostate_variant("1") == "std::string"

    assert m.cast_monostate_variant() == (None, 5, "Hello")

    assert (
        doc(m.load_monostate_variant)
        == "load_monostate_variant(arg0: None | typing.SupportsInt | str) -> str"
    )


def test_vec_of_reference_wrapper():
    """#171: Can't return reference wrappers (or STL structures containing them)"""
    assert (
        str(m.return_vec_of_reference_wrapper(UserType(4)))
        == "[UserType(1), UserType(2), UserType(3), UserType(4)]"
    )


def test_stl_pass_by_pointer(msg):
    """Passing nullptr or None to an STL container pointer is not expected to work"""
    with pytest.raises(TypeError) as excinfo:
        m.stl_pass_by_pointer()  # default value is `nullptr`
    assert (
        msg(excinfo.value)
        == """
        stl_pass_by_pointer(): incompatible function arguments. The following argument types are supported:
            1. (v: collections.abc.Sequence[typing.SupportsInt] = None) -> list[int]

        Invoked with:
    """
    )

    with pytest.raises(TypeError) as excinfo:
        m.stl_pass_by_pointer(None)
    assert (
        msg(excinfo.value)
        == """
        stl_pass_by_pointer(): incompatible function arguments. The following argument types are supported:
            1. (v: collections.abc.Sequence[typing.SupportsInt] = None) -> list[int]

        Invoked with: None
    """
    )

    assert m.stl_pass_by_pointer([1, 2, 3]) == [1, 2, 3]


def test_missing_header_message():
    """Trying convert `list` to a `std::vector`, or vice versa, without including
    <pybind11/stl.h> should result in a helpful suggestion in the error message"""
    import pybind11_cross_module_tests as cm

    expected_message = (
        "Did you forget to `#include <pybind11/stl.h>`? Or <pybind11/complex.h>,\n"
        "<pybind11/functional.h>, <pybind11/chrono.h>, etc. Some automatic\n"
        "conversions are optional and require extra headers to be included\n"
        "when compiling your pybind11 module."
    )

    with pytest.raises(TypeError) as excinfo:
        cm.missing_header_arg([1.0, 2.0, 3.0])
    assert expected_message in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        cm.missing_header_return()
    assert expected_message in str(excinfo.value)


def test_function_with_string_and_vector_string_arg():
    """Check if a string is NOT implicitly converted to a list, which was the
    behavior before fix of issue #1258"""
    assert m.func_with_string_or_vector_string_arg_overload(("A", "B")) == 2
    assert m.func_with_string_or_vector_string_arg_overload(["A", "B"]) == 2
    assert m.func_with_string_or_vector_string_arg_overload("A") == 3


@pytest.mark.skipif("env.GRAALPY", reason="Cannot reliably trigger GC")
def test_stl_ownership():
    cstats = ConstructorStats.get(m.Placeholder)
    assert cstats.alive() == 0
    r = m.test_stl_ownership()
    assert len(r) == 1
    del r
    assert cstats.alive() == 0


def test_array_cast_sequence():
    assert m.array_cast_sequence((1, 2, 3)) == [1, 2, 3]


def test_issue_1561():
    """check fix for issue #1561"""
    bar = m.Issue1561Outer()
    bar.list = [m.Issue1561Inner("bar")]
    assert bar.list
    assert bar.list[0].data == "bar"


def test_return_vector_bool_raw_ptr():
    # Add `while True:` for manual leak checking.
    v = m.return_vector_bool_raw_ptr()
    assert isinstance(v, list)
    assert len(v) == 4513


@pytest.mark.parametrize(
    ("fn", "offset"), [(m.pass_std_vector_int, 0), (m.pass_std_array_int_2, 1)]
)
def test_pass_std_vector_int(fn, offset):
    assert fn([7, 13]) == 140 + offset
    assert fn({6, 2}) == 116 + offset
    assert fn({"x": 8, "y": 11}.values()) == 138 + offset
    assert fn({3: None, 9: None}.keys()) == 124 + offset
    assert fn(i for i in [4, 17]) == 142 + offset
    assert fn(map(lambda i: i * 3, [8, 7])) == 190 + offset  # noqa: C417
    with pytest.raises(TypeError):
        fn({"x": 0, "y": 1})
    with pytest.raises(TypeError):
        fn({})


def test_pass_std_vector_pair_int():
    fn = m.pass_std_vector_pair_int
    assert fn({1: 2, 3: 4}.items()) == 406
    assert fn(zip([5, 17], [13, 9])) == 2222


def test_list_caster_fully_consumes_generator_object():
    def gen_invalid():
        yield from [1, 2.0, 3]

    gen_obj = gen_invalid()
    with pytest.raises(TypeError):
        m.pass_std_vector_int(gen_obj)
    assert not tuple(gen_obj)


def test_pass_std_set_int():
    fn = m.pass_std_set_int
    assert fn({3, 15}) == 254
    assert fn({5: None, 12: None}.keys()) == 251
    with pytest.raises(TypeError):
        fn([])
    with pytest.raises(TypeError):
        fn({})
    with pytest.raises(TypeError):
        fn({}.values())
    with pytest.raises(TypeError):
        fn(i for i in [])


def test_set_caster_dict_keys_failure():
    dict_keys = {1: None, 2.0: None, 3: None}.keys()
    # The asserts does not really exercise anything in pybind11, but if one of
    # them fails in some future version of Python, the set_caster load
    # implementation may need to be revisited.
    assert tuple(dict_keys) == (1, 2.0, 3)
    assert tuple(dict_keys) == (1, 2.0, 3)
    with pytest.raises(TypeError):
        m.pass_std_set_int(dict_keys)
    assert tuple(dict_keys) == (1, 2.0, 3)


class FakePyMappingMissingItems:
    def __getitem__(self, _):
        raise RuntimeError("Not expected to be called.")


class FakePyMappingWithItems(FakePyMappingMissingItems):
    def items(self):
        return ((1, 3), (2, 4))


class FakePyMappingBadItems(FakePyMappingMissingItems):
    def items(self):
        return ((1, 2), (3, "x"))


class FakePyMappingItemsNotCallable(FakePyMappingMissingItems):
    @property
    def items(self):
        return ((1, 2), (3, 4))


class FakePyMappingItemsWithArg(FakePyMappingMissingItems):
    def items(self, _):
        return ((1, 2), (3, 4))


class FakePyMappingGenObj(FakePyMappingMissingItems):
    def __init__(self, gen_obj):
        super().__init__()
        self.gen_obj = gen_obj

    def items(self):
        yield from self.gen_obj


def test_pass_std_map_int():
    fn = m.pass_std_map_int
    assert fn({1: 2, 3: 4}) == 4506
    with pytest.raises(TypeError):
        fn([])
    assert fn(FakePyMappingWithItems()) == 3507
    with pytest.raises(TypeError):
        fn(FakePyMappingMissingItems())
    with pytest.raises(TypeError):
        fn(FakePyMappingBadItems())
    with pytest.raises(TypeError):
        fn(FakePyMappingItemsNotCallable())
    with pytest.raises(TypeError):
        fn(FakePyMappingItemsWithArg())


@pytest.mark.parametrize(
    ("items", "expected_exception"),
    [
        (((1, 2), (3, "x"), (4, 5)), TypeError),
        (((1, 2), (3, 4, 5), (6, 7)), ValueError),
    ],
)
def test_map_caster_fully_consumes_generator_object(items, expected_exception):
    def gen_invalid():
        yield from items

    gen_obj = gen_invalid()
    with pytest.raises(expected_exception):
        m.pass_std_map_int(FakePyMappingGenObj(gen_obj))
    assert not tuple(gen_obj)


def test_sequence_caster_protocol(doc):
    from collections.abc import Sequence

    # Implements the Sequence protocol without explicitly inheriting from collections.abc.Sequence.
    class BareSequenceLike:
        def __init__(self, *args):
            self.data = tuple(args)

        def __len__(self):
            return len(self.data)

        def __getitem__(self, index):
            return self.data[index]

    # Implements the Sequence protocol by reusing BareSequenceLike's implementation.
    # Additionally, inherits from collections.abc.Sequence.
    class FormalSequenceLike(BareSequenceLike, Sequence):
        pass

    # convert mode
    assert (
        doc(m.roundtrip_std_vector_int)
        == "roundtrip_std_vector_int(arg0: collections.abc.Sequence[typing.SupportsInt]) -> list[int]"
    )
    assert m.roundtrip_std_vector_int([1, 2, 3]) == [1, 2, 3]
    assert m.roundtrip_std_vector_int((1, 2, 3)) == [1, 2, 3]
    assert m.roundtrip_std_vector_int(FormalSequenceLike(1, 2, 3)) == [1, 2, 3]
    assert m.roundtrip_std_vector_int(BareSequenceLike(1, 2, 3)) == [1, 2, 3]
    assert m.roundtrip_std_vector_int([]) == []
    assert m.roundtrip_std_vector_int(()) == []
    assert m.roundtrip_std_vector_int(BareSequenceLike()) == []
    # noconvert mode
    assert (
        doc(m.roundtrip_std_vector_int_noconvert)
        == "roundtrip_std_vector_int_noconvert(v: list[int]) -> list[int]"
    )
    assert m.roundtrip_std_vector_int_noconvert([1, 2, 3]) == [1, 2, 3]
    assert m.roundtrip_std_vector_int_noconvert((1, 2, 3)) == [1, 2, 3]
    assert m.roundtrip_std_vector_int_noconvert(FormalSequenceLike(1, 2, 3)) == [
        1,
        2,
        3,
    ]
    assert m.roundtrip_std_vector_int_noconvert(BareSequenceLike(1, 2, 3)) == [1, 2, 3]
    assert m.roundtrip_std_vector_int_noconvert([]) == []
    assert m.roundtrip_std_vector_int_noconvert(()) == []
    assert m.roundtrip_std_vector_int_noconvert(BareSequenceLike()) == []


def test_mapping_caster_protocol(doc):
    from collections.abc import Mapping

    # Implements the Mapping protocol without explicitly inheriting from collections.abc.Mapping.
    class BareMappingLike:
        def __init__(self, **kwargs):
            self.data = dict(kwargs)

        def __len__(self):
            return len(self.data)

        def __getitem__(self, key):
            return self.data[key]

        def __iter__(self):
            yield from self.data

    # Implements the Mapping protocol by reusing BareMappingLike's implementation.
    # Additionally, inherits from collections.abc.Mapping.
    class FormalMappingLike(BareMappingLike, Mapping):
        pass

    a1b2c3 = {"a": 1, "b": 2, "c": 3}
    # convert mode
    assert (
        doc(m.roundtrip_std_map_str_int)
        == "roundtrip_std_map_str_int(arg0: collections.abc.Mapping[str, typing.SupportsInt]) -> dict[str, int]"
    )
    assert m.roundtrip_std_map_str_int(a1b2c3) == a1b2c3
    assert m.roundtrip_std_map_str_int(FormalMappingLike(**a1b2c3)) == a1b2c3
    assert m.roundtrip_std_map_str_int({}) == {}
    assert m.roundtrip_std_map_str_int(FormalMappingLike()) == {}
    with pytest.raises(TypeError):
        m.roundtrip_std_map_str_int(BareMappingLike(**a1b2c3))
    # noconvert mode
    assert (
        doc(m.roundtrip_std_map_str_int_noconvert)
        == "roundtrip_std_map_str_int_noconvert(m: dict[str, int]) -> dict[str, int]"
    )
    assert m.roundtrip_std_map_str_int_noconvert(a1b2c3) == a1b2c3
    assert m.roundtrip_std_map_str_int_noconvert({}) == {}
    with pytest.raises(TypeError):
        m.roundtrip_std_map_str_int_noconvert(FormalMappingLike(**a1b2c3))
    with pytest.raises(TypeError):
        m.roundtrip_std_map_str_int_noconvert(BareMappingLike(**a1b2c3))


def test_set_caster_protocol(doc):
    from collections.abc import Set

    # Implements the Set protocol without explicitly inheriting from collections.abc.Set.
    class BareSetLike:
        def __init__(self, *args):
            self.data = set(args)

        def __len__(self):
            return len(self.data)

        def __contains__(self, item):
            return item in self.data

        def __iter__(self):
            yield from self.data

    # Implements the Set protocol by reusing BareSetLike's implementation.
    # Additionally, inherits from collections.abc.Set.
    class FormalSetLike(BareSetLike, Set):
        pass

    # convert mode
    assert (
        doc(m.roundtrip_std_set_int)
        == "roundtrip_std_set_int(arg0: collections.abc.Set[typing.SupportsInt]) -> set[int]"
    )
    assert m.roundtrip_std_set_int({1, 2, 3}) == {1, 2, 3}
    assert m.roundtrip_std_set_int(FormalSetLike(1, 2, 3)) == {1, 2, 3}
    assert m.roundtrip_std_set_int(set()) == set()
    assert m.roundtrip_std_set_int(FormalSetLike()) == set()
    with pytest.raises(TypeError):
        m.roundtrip_std_set_int(BareSetLike(1, 2, 3))
    # noconvert mode
    assert (
        doc(m.roundtrip_std_set_int_noconvert)
        == "roundtrip_std_set_int_noconvert(s: set[int]) -> set[int]"
    )
    assert m.roundtrip_std_set_int_noconvert({1, 2, 3}) == {1, 2, 3}
    assert m.roundtrip_std_set_int_noconvert(set()) == set()
    with pytest.raises(TypeError):
        m.roundtrip_std_set_int_noconvert(FormalSetLike(1, 2, 3))
    with pytest.raises(TypeError):
        m.roundtrip_std_set_int_noconvert(BareSetLike(1, 2, 3))



import json
import os
from os.path import exists, dirname, join
import subprocess
import sys

import numpy
import pybind11_stubgen

def main():
    import sys
    import os
    import subprocess
    from os.path import pardir, sep
    sys.path.append(os.getcwd())

    # make dir pyclipr
    os.makedirs('pyclipr', exist_ok=True)
    # copy pyclipr.cpython-312-darwin.so into pyclipr
    #subprocess.run(["ls"])

    # find the python shared object file
    # if os is mac or linux, find the file that starts with 'pyclipr.cpython-'
    if sys.platform.startswith('darwin') or sys.platform.startswith('linux'):
        pyclipr_lib = next((f for f in os.listdir(os.getcwd()) if f .startswith('pyclipr.cpython-') ), None)
    elif sys.platform.startswith('win32'):
        pyclipr_lib = next((f for f in os.listdir(os.getcwd() + '/release') if f .startswith('pyclipr.cp') and f.endswith('.pyd')), None)
        pyclipr_lib = './release/' + pyclipr_lib
    else:
        raise Exception('Unsupported platform')

    subprocess.run(['cp', pyclipr_lib, 'pyclipr'])
    # copy __init__.py into pyclipr folder
    subprocess.run(['cp', '__init__.py', 'pyclipr'])

    # run the process
    #subprocess.run(['pybind11-stubgen','--output-dir',  os.getcwd(), 'pyclipr'])
    sys.argv = [
        "<dummy>",
        "--output-dir",
        os.getcwd(),
        "pyclipr",
    ]

    pybind11_stubgen.main()


if __name__ == "__main__":
    main()


from __future__ import annotations

import pytest

from pybind11_tests import class_sh_disowning as m


def is_disowned(obj):
    try:
        obj.get()
    except ValueError:
        return True
    return False


def test_same_twice():
    while True:
        obj1a = m.Atype1(57)
        obj1b = m.Atype1(62)
        assert m.same_twice(obj1a, obj1b) == (57 * 10 + 1) * 100 + (62 * 10 + 1) * 10
        assert is_disowned(obj1a)
        assert is_disowned(obj1b)
        obj1c = m.Atype1(0)
        with pytest.raises(ValueError):
            # Disowning works for one argument, but not both.
            m.same_twice(obj1c, obj1c)
        assert is_disowned(obj1c)
        return  # Comment out for manual leak checking (use `top` command).


def test_mixed():
    first_pass = True
    while True:
        obj1a = m.Atype1(90)
        obj2a = m.Atype2(25)
        assert m.mixed(obj1a, obj2a) == (90 * 10 + 1) * 200 + (25 * 10 + 2) * 20
        assert is_disowned(obj1a)
        assert is_disowned(obj2a)

        # The C++ order of evaluation of function arguments is (unfortunately) unspecified:
        # https://en.cppreference.com/w/cpp/language/eval_order
        # Read on.
        obj1b = m.Atype1(0)
        with pytest.raises(ValueError):
            # If the 1st argument is evaluated first, obj1b is disowned before the conversion for
            # the already disowned obj2a fails as expected.
            m.mixed(obj1b, obj2a)
        obj2b = m.Atype2(0)
        with pytest.raises(ValueError):
            # If the 2nd argument is evaluated first, obj2b is disowned before the conversion for
            # the already disowned obj1a fails as expected.
            m.mixed(obj1a, obj2b)

        # Either obj1b or obj2b was disowned in the expected failed m.mixed() calls above, but not
        # both.
        is_disowned_results = (is_disowned(obj1b), is_disowned(obj2b))
        assert is_disowned_results.count(True) == 1
        if first_pass:
            first_pass = False
            ix = is_disowned_results.index(True) + 1
            print(f"\nC++ function argument {ix} is evaluated first.")

        return  # Comment out for manual leak checking (use `top` command).


def test_overloaded():
    while True:
        obj1 = m.Atype1(81)
        obj2 = m.Atype2(60)
        with pytest.raises(TypeError):
            m.overloaded(obj1, "NotInt")
        assert obj1.get() == 81 * 10 + 1  # Not disowned.
        assert m.overloaded(obj1, 3) == (81 * 10 + 1) * 30 + 3
        with pytest.raises(TypeError):
            m.overloaded(obj2, "NotInt")
        assert obj2.get() == 60 * 10 + 2  # Not disowned.
        assert m.overloaded(obj2, 2) == (60 * 10 + 2) * 40 + 2
        return  # Comment out for manual leak checking (use `top` command).



from __future__ import annotations

import enum
import pickle

import pytest

import env
from pybind11_tests import native_enum as m

SMALLENUM_MEMBERS = (
    ("a", 0),
    ("b", 1),
    ("c", 2),
)

COLOR_MEMBERS = (
    ("red", 0),
    ("yellow", 1),
    ("green", 20),
    ("blue", 21),
)

ALTITUDE_MEMBERS = (
    ("high", "h"),
    ("low", "l"),
)

FLAGS_UCHAR_MEMBERS = (
    ("bit0", 0x1),
    ("bit1", 0x2),
    ("bit2", 0x4),
)

FLAGS_UINT_MEMBERS = (
    ("bit0", 0x1),
    ("bit1", 0x2),
    ("bit2", 0x4),
)

CLASS_WITH_ENUM_IN_CLASS_MEMBERS = (
    ("one", 0),
    ("two", 1),
)

EXPORT_VALUES_MEMBERS = (
    ("exv0", 0),
    ("exv1", 1),
)

MEMBER_DOC_MEMBERS = (
    ("mem0", 0),
    ("mem1", 1),
    ("mem2", 2),
)

FUNC_SIG_RENDERING_MEMBERS = ()

ENUM_TYPES_AND_MEMBERS = (
    (m.smallenum, SMALLENUM_MEMBERS),
    (m.color, COLOR_MEMBERS),
    (m.altitude, ALTITUDE_MEMBERS),
    (m.flags_uchar, FLAGS_UCHAR_MEMBERS),
    (m.flags_uint, FLAGS_UINT_MEMBERS),
    (m.export_values, EXPORT_VALUES_MEMBERS),
    (m.member_doc, MEMBER_DOC_MEMBERS),
    (m.func_sig_rendering, FUNC_SIG_RENDERING_MEMBERS),
    (m.class_with_enum.in_class, CLASS_WITH_ENUM_IN_CLASS_MEMBERS),
)

ENUM_TYPES = [_[0] for _ in ENUM_TYPES_AND_MEMBERS]


@pytest.mark.parametrize("enum_type", ENUM_TYPES)
def test_enum_type(enum_type):
    assert isinstance(enum_type, enum.EnumMeta)
    assert enum_type.__module__ == m.__name__


@pytest.mark.parametrize(("enum_type", "members"), ENUM_TYPES_AND_MEMBERS)
def test_enum_members(enum_type, members):
    for name, value in members:
        assert enum_type[name].value == value


@pytest.mark.parametrize(("enum_type", "members"), ENUM_TYPES_AND_MEMBERS)
def test_pickle_roundtrip(enum_type, members):
    for name, _ in members:
        orig = enum_type[name]
        # This only works if __module__ is correct.
        serialized = pickle.dumps(orig)
        restored = pickle.loads(serialized)
        assert restored == orig


@pytest.mark.parametrize("enum_type", [m.flags_uchar, m.flags_uint])
def test_enum_flag(enum_type):
    bits02 = enum_type.bit0 | enum_type.bit2
    assert enum_type.bit0 in bits02
    assert enum_type.bit1 not in bits02
    assert enum_type.bit2 in bits02


def test_export_values():
    assert m.exv0 is m.export_values.exv0
    assert m.exv1 is m.export_values.exv1


def test_class_doc():
    pure_native = enum.IntEnum("pure_native", (("mem", 0),))
    assert m.smallenum.__doc__ == "doc smallenum"
    assert m.color.__doc__ == pure_native.__doc__


def test_member_doc():
    pure_native = enum.IntEnum("pure_native", (("mem", 0),))
    assert m.member_doc.mem0.__doc__ == "docA"
    assert m.member_doc.mem1.__doc__ == pure_native.mem.__doc__
    assert m.member_doc.mem2.__doc__ == "docC"


def test_pybind11_isinstance_color():
    for name, _ in COLOR_MEMBERS:
        assert m.isinstance_color(m.color[name])
    assert not m.isinstance_color(m.color)
    for name, _ in SMALLENUM_MEMBERS:
        assert not m.isinstance_color(m.smallenum[name])
    assert not m.isinstance_color(m.smallenum)
    assert not m.isinstance_color(None)


def test_pass_color_success():
    for name, value in COLOR_MEMBERS:
        assert m.pass_color(m.color[name]) == value


def test_pass_color_fail():
    with pytest.raises(TypeError) as excinfo:
        m.pass_color(None)
    assert "pybind11_tests.native_enum.color" in str(excinfo.value)


def test_return_color_success():
    for name, value in COLOR_MEMBERS:
        assert m.return_color(value) == m.color[name]


def test_return_color_fail():
    with pytest.raises(ValueError) as excinfo_direct:
        m.color(2)
    with pytest.raises(ValueError) as excinfo_cast:
        m.return_color(2)
    assert str(excinfo_cast.value) == str(excinfo_direct.value)


def test_return_color_ptr():
    assert m.return_color_const_ptr() == m.color.red
    assert m.return_color_mutbl_ptr() == m.color.green


def test_property_type_hint():
    prop = m.class_with_enum.__dict__["nested_value"]
    assert isinstance(prop, property)
    assert prop.fget.__doc__.startswith(
        "(self: pybind11_tests.native_enum.class_with_enum)"
        " -> pybind11_tests.native_enum.class_with_enum.in_class"
    )


def test_func_sig_rendering():
    assert m.pass_and_return_func_sig_rendering.__doc__.startswith(
        "pass_and_return_func_sig_rendering(e: pybind11_tests.native_enum.func_sig_rendering)"
        " -> pybind11_tests.native_enum.func_sig_rendering"
    )


def test_type_caster_enum_type_enabled_false():
    # This is really only a "does it compile" test.
    assert m.pass_some_proto_enum(None) is None
    assert m.return_some_proto_enum() is None


@pytest.mark.skipif(isinstance(m.obj_cast_color_ptr, str), reason=m.obj_cast_color_ptr)
def test_obj_cast_color_ptr():
    with pytest.raises(RuntimeError) as excinfo:
        m.obj_cast_color_ptr(m.color.red)
    assert str(excinfo.value) == "Unable to cast native enum type to reference"


def test_py_cast_color_handle():
    for name, value in COLOR_MEMBERS:
        assert m.py_cast_color_handle(m.color[name]) == value


def test_exercise_import_or_getattr_leading_dot():
    with pytest.raises(ValueError) as excinfo:
        m.exercise_import_or_getattr(m, ".")
    assert str(excinfo.value) == "Invalid fully-qualified name `.` (native_type_name)"


def test_exercise_import_or_getattr_bad_top_level():
    with pytest.raises(ImportError) as excinfo:
        m.exercise_import_or_getattr(m, "NeVeRLaNd")
    assert (
        str(excinfo.value)
        == "Failed to import top-level module `NeVeRLaNd` (native_type_name)"
    )


def test_exercise_import_or_getattr_dot_dot():
    with pytest.raises(ValueError) as excinfo:
        m.exercise_import_or_getattr(m, "enum..")
    assert (
        str(excinfo.value) == "Invalid fully-qualified name `enum..` (native_type_name)"
    )


def test_exercise_import_or_getattr_bad_enum_attr():
    with pytest.raises(ImportError) as excinfo:
        m.exercise_import_or_getattr(m, "enum.NoNeXiStInG")
    lines = str(excinfo.value).splitlines()
    assert len(lines) >= 5
    assert (
        lines[0]
        == "Failed to import or getattr `NoNeXiStInG` from `enum` (native_type_name)"
    )
    assert lines[1] == "-------- getattr exception --------"
    ix = lines.index("-------- import exception --------")
    assert ix >= 3
    assert len(lines) > ix + 0


def test_native_enum_data_missing_finalize_error_message():
    msg = m.native_enum_data_missing_finalize_error_message("Fake")
    assert msg == 'pybind11::native_enum<...>("Fake", ...): MISSING .finalize()'


@pytest.mark.parametrize(
    "func", [m.native_enum_ctor_malformed_utf8, m.native_enum_value_malformed_utf8]
)
def test_native_enum_malformed_utf8(func):
    if env.GRAALPY and func is m.native_enum_ctor_malformed_utf8:
        pytest.skip("GraalPy does not raise UnicodeDecodeError")
    malformed_utf8 = b"\x80"
    with pytest.raises(UnicodeDecodeError):
        func(malformed_utf8)


def test_native_enum_double_finalize():
    with pytest.raises(RuntimeError) as excinfo:
        m.native_enum_double_finalize(m)
    assert (
        str(excinfo.value)
        == 'pybind11::native_enum<...>("fake_native_enum_double_finalize"): DOUBLE finalize'
    )


def test_native_enum_value_after_finalize():
    with pytest.raises(RuntimeError) as excinfo:
        m.native_enum_value_after_finalize(m)
    assert (
        str(excinfo.value)
        == 'pybind11::native_enum<...>("fake_native_enum_value_after_finalize"): value after finalize'
    )


def test_double_registration_native_enum():
    with pytest.raises(RuntimeError) as excinfo:
        m.double_registration_native_enum(m)
    assert (
        str(excinfo.value)
        == 'pybind11::native_enum<...>("fake_double_registration_native_enum") is already registered!'
    )


def test_native_enum_name_clash():
    m.fake_native_enum_name_clash = None
    with pytest.raises(RuntimeError) as excinfo:
        m.native_enum_name_clash(m)
    assert (
        str(excinfo.value)
        == 'pybind11::native_enum<...>("fake_native_enum_name_clash"):'
        " an object with that name is already defined"
    )


def test_native_enum_value_name_clash():
    m.fake_native_enum_value_name_clash_x = None
    with pytest.raises(RuntimeError) as excinfo:
        m.native_enum_value_name_clash(m)
    assert (
        str(excinfo.value)
        == 'pybind11::native_enum<...>("fake_native_enum_value_name_clash")'
        '.value("fake_native_enum_value_name_clash_x"):'
        " an object with that name is already defined"
    )


def test_double_registration_enum_before_native_enum():
    with pytest.raises(RuntimeError) as excinfo:
        m.double_registration_enum_before_native_enum(m)
    assert (
        str(excinfo.value)
        == 'pybind11::native_enum<...>("fake_enum_first") is already registered'
        " as a `pybind11::enum_` or `pybind11::class_`!"
    )


def test_double_registration_native_enum_before_enum():
    with pytest.raises(RuntimeError) as excinfo:
        m.double_registration_native_enum_before_enum(m)
    assert (
        str(excinfo.value)
        == 'pybind11::enum_ "name_must_be_different_to_reach_desired_code_path"'
        " is already registered as a pybind11::native_enum!"
    )


def test_native_enum_missing_finalize_failure():
    if not isinstance(m.native_enum_missing_finalize_failure, str):
        m.native_enum_missing_finalize_failure()
        pytest.fail("Process termination expected.")



from __future__ import annotations

import pytest

m = pytest.importorskip("pybind11_tests.constants_and_functions")


def test_constants():
    assert m.some_constant == 14


def test_function_overloading():
    assert m.test_function() == "test_function()"
    assert m.test_function(7) == "test_function(7)"
    assert m.test_function(m.MyEnum.EFirstEntry) == "test_function(enum=1)"
    assert m.test_function(m.MyEnum.ESecondEntry) == "test_function(enum=2)"

    assert m.test_function() == "test_function()"
    assert m.test_function("abcd") == "test_function(char *)"
    assert m.test_function(1, 1.0) == "test_function(int, float)"
    assert m.test_function(1, 1.0) == "test_function(int, float)"
    assert m.test_function(2.0, 2) == "test_function(float, int)"


def test_bytes():
    assert m.print_bytes(m.return_bytes()) == "bytes[1 0 2 0]"


def test_exception_specifiers():
    c = m.C()
    assert c.m1(2) == 1
    assert c.m2(3) == 1
    assert c.m3(5) == 2
    assert c.m4(7) == 3
    assert c.m5(10) == 5
    assert c.m6(14) == 8
    assert c.m7(20) == 13
    assert c.m8(29) == 21

    assert m.f1(33) == 34
    assert m.f2(53) == 55
    assert m.f3(86) == 89
    assert m.f4(140) == 144


def test_function_record_leaks():
    class RaisingRepr:
        def __repr__(self):
            raise RuntimeError("Surprise!")

    with pytest.raises(RuntimeError):
        m.register_large_capture_with_invalid_arguments(m)
    with pytest.raises(RuntimeError):
        m.register_with_raising_repr(m, RaisingRepr())


def test_noexcept_lambda():
    assert m.l1() == 0



import matplotlib.pyplot as plt
import numpy as np

import pyclipr

# Tuple definition of a path
path = np.array([(0.0, 0.), (0, 105.1234), (100, 105.1234), (100, 0), (0, 0)])
path2 = np.array([(10.0, 10.0), (1.0, 50), (50.0, 50), (50., 10.0), (10.0, 10.0)])
path2 = np.flipud(path2)

path3 = np.array([(-10.0, -10.0), (70.0, -10), (70.0, 70), (-10., 70.0), (-10.0, -10.0)])

"""
Create an offsetting PyClipr tool
"""
po = pyclipr.ClipperOffset()

# Set the scale factor to convert to internal integer representation
# Note the default is set to 1000 units
po.scaleFactor = int(1e4)

# add the path - ensuring to use Polygon for the endType argument
po.addPaths([path, path2], pyclipr.JoinType.Miter, pyclipr.EndType.Polygon)
"""
Apply the offsetting operation using a delta/offset
NoteL this automatically scaled interally scaled within pyclipr.ClipperOffset
"""
offsetOut = po.execute(-1.0)

import pyslm.visualise
handle = pyslm.visualise.plotPolygon([path, path2])
pyslm.visualise.plotPolygon(offsetOut, handle=handle, lineColor='r')


"""
Create a Pyclipr clipping tool
"""
pc = pyclipr.Clipper()
pc.scaleFactor = int(1000)

"""
Add the paths to the clipping object. Ensure the subject and clip arguments are set to differentiate
the paths during the Boolean operation. The final argument specifies if the path is open.
"""
pc.addPaths(offsetOut, pyclipr.Subject)

# Note it is possible to add a path as a numpy array (nx2) or as a list of tuples
pc.addPath(np.array(path3), pyclipr.Clip)

"""
Perform Polygon Clipping 
The fill rule is optional and by defualt in ClipperLib is set to EvenOdd
"""

out = pc.execute(pyclipr.Intersection)
out2 = pc.execute(pyclipr.Union, pyclipr.FillRule.EvenOdd)
out3 = pc.execute(pyclipr.Difference, pyclipr.FillRule.EvenOdd)
out4 = pc.execute(pyclipr.Xor, pyclipr.FillRule.EvenOdd)

"""
Returning PolyTree structures following clipping operations:
Using execute2 returns a PolyTree structure that provides hierarchical information
"""
# if the paths are interior or exterior
outB = pc.execute2(pyclipr.Intersection)

# An alternative equivalent name is executeTree
outB = pc.executeTree(pyclipr.Intersection)

"""
Test Open Path Clipping:
# Pyclipr can be used for clipping open paths. This has become simple to complete using the Clipper2 library
"""

# Create a new Clipper Object
pc2 = pyclipr.Clipper()
pc2.scaleFactor = int(1e5)

"""
The open path is added as a subject (note the final argument 'isOpen' set to True). By default all paths are 
assumed to be closed in pyclipr
"""
pc2.addPath(((40, -10), (50, 130)), pyclipr.Subject, True)

# The clipping object is usually set to the Polygon
pc2.addPaths(offsetOut, pyclipr.Clip, False)

""" 
Test the return types for open path clipping with option enabled. When The returnOpenPaths argument is set to True
this will return the open paths. Note this function only works well using the Boolean intersection option
"""
outC = pc2.execute(pyclipr.Intersection, pyclipr.FillRule.NonZero)
outC2, openPathsC = pc2.execute(pyclipr.Intersection, pyclipr.FillRule.NonZero, returnOpenPaths=True)

outD = pc2.execute2(pyclipr.Intersection, pyclipr.FillRule.NonZero)
outD2, openPathsD = pc2.execute2(pyclipr.Intersection, pyclipr.FillRule.NonZero, returnOpenPaths=True)

"""
Plot the results from the examples
"""
pathPoly = np.array(path)

plt.figure()
plt.axis('equal')

# Plot the original polygon
plt.fill(pathPoly[:, 0], pathPoly[:, 1], 'b', alpha=0.1, linewidth=1.0, linestyle='dashed', edgecolor='#000')

# Plot the offset square
plt.fill(offsetOut[0][:, 0], offsetOut[0][:, 1],
         linewidth=1.0, linestyle='dashed', edgecolor='#333', facecolor='none')

# Plot the intersection
plt.fill(out[0][:, 0], out[0][:, 1], facecolor='#75507b')

# Plot the open path intersection
plt.plot(openPathsC[0][:, 0], openPathsC[0][:, 1],
         color='#222', linewidth=1.0, linestyle='dashed', marker='.', markersize=20.0)


"""Helper script to download source archives and upload them to the Eigen GitLab generic package registry."""

import os
import requests
import hashlib
import argparse
import sys
import tempfile

EIGEN_PROJECT_ID = 15462818  # Taken from the gitlab project page.


def calculate_sha256(filepath: str):
    """Calculates the SHA256 checksum of a file."""
    sha256_hash = hashlib.sha256()
    with open(filepath, "rb") as f:
        # Read and update hash in chunks of 4K
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def upload_to_generic_registry(
    gitlab_private_token: str, package_name: str, package_version: str, filepath: str
):
    """Uploads a file to the GitLab generic package registry."""
    headers = {"PRIVATE-TOKEN": gitlab_private_token}
    filename = os.path.basename(filepath)
    upload_url = f"https://gitlab.com/api/v4/projects/{EIGEN_PROJECT_ID}/packages/generic/{package_name}/{package_version}/{filename}"

    print(f"Uploading {filename} to {upload_url}...")
    try:
        with open(filepath, "rb") as f:
            response = requests.put(upload_url, headers=headers, data=f)
            response.raise_for_status()
        print(f"Successfully uploaded {filename}.")
        return True
    except requests.exceptions.RequestException as e:
        print(f"Error uploading {filename}: {e}")
        if e.response is not None:
            print(f"Response content: {e.response.text}")
        return False


def main():
    """Main function to download archives and upload them to the registry."""
    parser = argparse.ArgumentParser(
        description="Download GitLab release archives for Eigen and upload them to the generic package registry."
    )
    parser.add_argument(
        "--gitlab_private_token",
        type=str,
        help="GitLab private API token.  Defaults to the GITLAB_PRIVATE_TOKEN environment variable if set.",
    )
    parser.add_argument(
        "--version",
        required=True,
        help="Specify a single version (tag name) to process.",
    )
    parser.add_argument(
        "--download-dir", help=f"Directory to store temporary downloads (optional)."
    )
    args = parser.parse_args()

    if not args.gitlab_private_token:
        args.gitlab_private_token = os.getenv("GITLAB_PRIVATE_TOKEN")
        if not args.gitlab_private_token:
            print("Could not determine GITLAB_PRIVATE_TOKEN.", file=sys.stderr)
            parser.print_usage()
            sys.exit(1)

    # Create download directory if it doesn't exist.
    cleanup_download_dir = False
    if args.download_dir:
        if not os.path.exists(args.download_dir):
            cleanup_download_dir = True
            os.makedirs(args.download_dir)
    else:
        args.download_dir = tempfile.mkdtemp()
        cleanup_download_dir = True

    for ext in ["tar.gz", "tar.bz2", "tar", "zip"]:
        archive_filename = f"eigen-{args.version}.{ext}"
        archive_url = f"https://gitlab.com/libeigen/eigen/-/archive/{args.version}/{archive_filename}"
        archive_filepath = os.path.join(args.download_dir, archive_filename)

        # Download the archive
        print(f"Downloading {archive_url}...")
        try:
            response = requests.get(archive_url, stream=True)
            response.raise_for_status()
            with open(archive_filepath, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)
            print(f"Downloaded to {archive_filepath}")
        except requests.exceptions.RequestException as e:
            print(f"Error downloading {archive_url}: {e}. Skipping.")
            continue

        # Calculate SHA256 sum
        sha256_sum = calculate_sha256(archive_filepath)
        print(f"SHA256 sum: {sha256_sum}")

        # Create SHA256 sum file
        sha_filename = f"{archive_filename}.sha256"
        sha_filepath = os.path.join(args.download_dir, sha_filename)
        with open(sha_filepath, "w") as f:
            f.write(f"{sha256_sum}  {archive_filename}\n")
        print(f"Created SHA256 file: {sha_filepath}")

        # Upload archive to generic registry
        if not upload_to_generic_registry(
            args.gitlab_private_token, "eigen", args.version, archive_filepath
        ):
            # If upload fails, clean up and move to the next release
            os.remove(archive_filepath)
            os.remove(sha_filepath)
            continue

        # Upload SHA256 sum file to generic registry
        upload_to_generic_registry(
            args.gitlab_private_token, "eigen", args.version, sha_filepath
        )

        # Clean up downloaded files
        print("Cleaning up local files...")
        os.remove(archive_filepath)
        os.remove(sha_filepath)

    # Clean up the download directory if it's empty
    if cleanup_download_dir and not os.listdir(args.download_dir):
        os.rmdir(args.download_dir)


if __name__ == "__main__":
    main()



from __future__ import annotations

import pytest

from pybind11_tests import class_sh_property_non_owning as m


@pytest.mark.parametrize("persistent_holder", [True, False])
@pytest.mark.parametrize(
    ("core_fld", "expected"),
    [
        ("core_fld_value_ro", (13, 24)),
        ("core_fld_value_rw", (13, 24)),
        ("core_fld_shared_ptr_ro", (14, 25)),
        ("core_fld_shared_ptr_rw", (14, 25)),
        ("core_fld_raw_ptr_ro", (14, 25)),
        ("core_fld_raw_ptr_rw", (14, 25)),
        ("core_fld_unique_ptr_rw", (15, 26)),
    ],
)
def test_core_fld_common(core_fld, expected, persistent_holder):
    if persistent_holder:
        h = m.DataFieldsHolder(2)
        for i, exp in enumerate(expected):
            c = getattr(h.vec_at(i), core_fld)
            assert c.int_value == exp
    else:
        for i, exp in enumerate(expected):
            c = getattr(m.DataFieldsHolder(2).vec_at(i), core_fld)
            assert c.int_value == exp



from __future__ import annotations

import pytest

import pybind11_tests.class_sh_trampoline_self_life_support as m


class PyBig5(m.Big5):
    pass


def test_m_big5():
    obj = m.Big5("Seed")
    assert obj.history == "Seed"
    o1, o2 = m.action(obj, 0)
    assert o1 is not obj
    assert o1.history == "Seed"
    with pytest.raises(ValueError) as excinfo:
        _ = obj.history
    assert "Python instance was disowned" in str(excinfo.value)
    assert o2 is None


@pytest.mark.parametrize(
    ("action_id", "expected_history"),
    [
        (0, "Seed_CpCtor"),
        (1, "Seed_MvCtor"),
        (2, "Seed_OpEqLv"),
        (3, "Seed_OpEqRv"),
    ],
)
def test_py_big5(action_id, expected_history):
    obj = PyBig5("Seed")
    assert obj.history == "Seed"
    o1, o2 = m.action(obj, action_id)
    assert o1 is obj
    assert o2.history == expected_history



from __future__ import annotations

import pytest

from pybind11_tests import class_sh_factory_constructors as m


def test_atyp_factories():
    assert m.atyp_valu().get_mtxt() == "Valu"
    assert m.atyp_rref().get_mtxt() == "Rref"
    # sert m.atyp_cref().get_mtxt() == "Cref"
    # sert m.atyp_mref().get_mtxt() == "Mref"
    # sert m.atyp_cptr().get_mtxt() == "Cptr"
    assert m.atyp_mptr().get_mtxt() == "Mptr"
    assert m.atyp_shmp().get_mtxt() == "Shmp"
    assert m.atyp_shcp().get_mtxt() == "Shcp"
    assert m.atyp_uqmp().get_mtxt() == "Uqmp"
    assert m.atyp_uqcp().get_mtxt() == "Uqcp"
    assert m.atyp_udmp().get_mtxt() == "Udmp"
    assert m.atyp_udcp().get_mtxt() == "Udcp"


@pytest.mark.parametrize(
    ("init_args", "expected"),
    [
        ((3,), 300),
        ((5, 7), 570),
        ((9, 11, 13), 1023),
    ],
)
def test_with_alias_success(init_args, expected):
    assert m.with_alias(*init_args).val == expected


@pytest.mark.parametrize(
    ("num_init_args", "smart_ptr"),
    [
        (4, "std::unique_ptr"),
        (5, "std::shared_ptr"),
    ],
)
def test_with_alias_invalid(num_init_args, smart_ptr):
    class PyDrvdWithAlias(m.with_alias):
        pass

    with pytest.raises(TypeError) as excinfo:
        PyDrvdWithAlias(*((0,) * num_init_args))
    assert (
        str(excinfo.value)
        == "pybind11::init(): construction failed: returned "
        + smart_ptr
        + " pointee is not an alias instance"
    )



import matplotlib.pyplot as plt
import numpy as np

import pyclipr

""" 
Test Open Path Clipping with Z-Attribute:
Construct a path (nx3 list) with an additional z-attribute. The Z-attribute is currently set as a floating point number
and the 'maximum' value across the intersecting edges is returned by default. In future releases, the z-attribute will
be customisable at computational expense. In this example, the single line has a z-value of 3.0 and 4.0, therefore,
the maximum Z value will be assigned to the new vertices in the resultant clipped line. 
"""
openPathPolyClipper = [(-1.0, -5.,0.1),
                       (-1.0, 105.1234, 0.1),
                       (100, 105.1234,0.1),
                       (100., -5.,0.1)]

""" Construct a Clipping Tool Object"""
pc2 = pyclipr.Clipper()
pc2.scaleFactor = int(1e6)

""" Add the existing nx3 path to the clipping tool """
pc2.addPath(openPathPolyClipper, pyclipr.Clip, False)

""" Add a second open path (line) to the clipping tool for intersection """
pc2.addPath( ((50,-20,3.0),(49.9,150,4.0)), pyclipr.Subject, isOpen = True)

""" Test the return types for open path clipping with different options selected """

# The first option will return only the paths without return the additional Z attribute.
# Essentially this will clip only the paths
closedPolys, openPaths, = pc2.execute(pyclipr.Intersection, pyclipr.FillRule.EvenOdd, returnOpenPaths=True)

# Including the returnZ=True argument will return both open and closed paths with the Z attribute
closedPolys, openPaths, closedPolyZVals, openPathZVals = pc2.execute(pyclipr.Intersection, pyclipr.FillRule.EvenOdd,
                                                                       returnOpenPaths=True, returnZ=True)

# Print the open path with the Z attribute included
for i, pnt in enumerate(openPaths[0]):
    print('Open Path with Z: x={:.3f}, y={:.3f} {:f}'.format(pnt[0], pnt[1], openPathZVals[0][i]))

""" Test the return types for open path clipping with option enabled """
outD = pc2.execute2(pyclipr.Intersection,  pyclipr.FillRule.NonZero)
outD2 = pc2.execute2(pyclipr.Intersection,  pyclipr.FillRule.NonZero, returnOpenPaths=True)
outD2withZ = pc2.execute2(pyclipr.Intersection,  pyclipr.FillRule.NonZero, returnOpenPaths=True, returnZ=True)


"""
Plot the results from the examples
"""
pathPoly = np.array(openPathPolyClipper)[:,:2]

plt.figure()
plt.axis('equal')

# Plot the original polygon
plt.fill(pathPoly[:, 0], pathPoly[:, 1], 'b', linewidth=1.0, linestyle='dashed', edgecolor='#333', facecolor='#b08fb6')

# Plot the open path intersection
plt.plot(openPaths[0][:, 0], openPaths[0][:, 1],
         color='#222', linewidth=1.0, linestyle='dashed', marker='.', markersize=20.0)


# The compact 4-character naming scheme (e.g. mptr, cptr, shcp) is explained at the top of
# test_class_sh_property.cpp.
from __future__ import annotations

import pytest

import env  # noqa: F401
from pybind11_tests import class_sh_property as m


@pytest.mark.skipif(
    "env.PYPY or env.GRAALPY", reason="gc after `del field` is apparently deferred"
)
@pytest.mark.parametrize("m_attr", ["m_valu_readonly", "m_valu_readwrite"])
def test_valu_getter(m_attr):
    # Reduced from PyCLIF test:
    # https://github.com/google/clif/blob/c371a6d4b28d25d53a16e6d2a6d97305fb1be25a/clif/testing/python/nested_fields_test.py#L56
    outer = m.Outer()
    field = getattr(outer, m_attr)
    assert field.num == -99
    with pytest.raises(ValueError) as excinfo:
        m.DisownOuter(outer)
    assert str(excinfo.value) == "Cannot disown use_count != 1 (load_as_unique_ptr)."
    del field
    m.DisownOuter(outer)
    with pytest.raises(ValueError, match="Python instance was disowned") as excinfo:
        getattr(outer, m_attr)


def test_valu_setter():
    outer = m.Outer()
    assert outer.m_valu_readonly.num == -99
    assert outer.m_valu_readwrite.num == -99
    field = m.Field()
    field.num = 35
    outer.m_valu_readwrite = field
    assert outer.m_valu_readonly.num == 35
    assert outer.m_valu_readwrite.num == 35


@pytest.mark.parametrize("m_attr", ["m_shmp", "m_shcp"])
def test_shp(m_attr):
    m_attr_readonly = m_attr + "_readonly"
    m_attr_readwrite = m_attr + "_readwrite"
    outer = m.Outer()
    assert getattr(outer, m_attr_readonly) is None
    assert getattr(outer, m_attr_readwrite) is None
    field = m.Field()
    field.num = 43
    setattr(outer, m_attr_readwrite, field)
    assert getattr(outer, m_attr_readonly).num == 43
    assert getattr(outer, m_attr_readwrite).num == 43
    getattr(outer, m_attr_readonly).num = 57
    getattr(outer, m_attr_readwrite).num = 57
    assert field.num == 57
    del field
    assert getattr(outer, m_attr_readonly).num == 57
    assert getattr(outer, m_attr_readwrite).num == 57


@pytest.mark.parametrize(
    ("field_type", "num_default", "outer_type"),
    [
        (m.ClassicField, -88, m.ClassicOuter),
        (m.Field, -99, m.Outer),
    ],
)
@pytest.mark.parametrize("m_attr", ["m_mptr", "m_cptr"])
@pytest.mark.parametrize("r_kind", ["_readonly", "_readwrite"])
def test_ptr(field_type, num_default, outer_type, m_attr, r_kind):
    m_attr_r_kind = m_attr + r_kind
    outer = outer_type()
    assert getattr(outer, m_attr_r_kind) is None
    field = field_type()
    assert field.num == num_default
    setattr(outer, m_attr + "_readwrite", field)
    assert getattr(outer, m_attr_r_kind).num == num_default
    field.num = 76
    assert getattr(outer, m_attr_r_kind).num == 76
    # Change to -88 or -99 to demonstrate Undefined Behavior (dangling pointer).
    if num_default == 88 and m_attr == "m_mptr":
        del field
    assert getattr(outer, m_attr_r_kind).num == 76


@pytest.mark.parametrize("m_attr_readwrite", ["m_uqmp_readwrite", "m_uqcp_readwrite"])
def test_uqp(m_attr_readwrite):
    outer = m.Outer()
    assert getattr(outer, m_attr_readwrite) is None
    field_orig = m.Field()
    field_orig.num = 39
    setattr(outer, m_attr_readwrite, field_orig)
    with pytest.raises(ValueError, match="Python instance was disowned"):
        _ = field_orig.num
    field_retr1 = getattr(outer, m_attr_readwrite)
    assert getattr(outer, m_attr_readwrite) is None
    assert field_retr1.num == 39
    field_retr1.num = 93
    setattr(outer, m_attr_readwrite, field_retr1)
    with pytest.raises(ValueError):
        _ = field_retr1.num
    field_retr2 = getattr(outer, m_attr_readwrite)
    assert field_retr2.num == 93


# Proof-of-concept (POC) for safe & intuitive Python access to unique_ptr members.
# The C++ member unique_ptr is disowned to a temporary Python object for accessing
# an attribute of the member. After the attribute was accessed, the Python object
# is disowned back to the C++ member unique_ptr.
# Productizing this POC is left for a future separate PR, as needed.
class unique_ptr_field_proxy_poc:
    def __init__(self, obj, field_name):
        object.__setattr__(self, "__obj", obj)
        object.__setattr__(self, "__field_name", field_name)

    def __getattr__(self, *args, **kwargs):
        return _proxy_dereference(self, getattr, *args, **kwargs)

    def __setattr__(self, *args, **kwargs):
        return _proxy_dereference(self, setattr, *args, **kwargs)

    def __delattr__(self, *args, **kwargs):
        return _proxy_dereference(self, delattr, *args, **kwargs)


def _proxy_dereference(proxy, xxxattr, *args, **kwargs):
    obj = object.__getattribute__(proxy, "__obj")
    field_name = object.__getattribute__(proxy, "__field_name")
    field = getattr(obj, field_name)  # Disowns the C++ unique_ptr member.
    assert field is not None
    try:
        return xxxattr(field, *args, **kwargs)
    finally:
        setattr(obj, field_name, field)  # Disowns the temporary Python object (field).


@pytest.mark.parametrize("m_attr", ["m_uqmp", "m_uqcp"])
def test_unique_ptr_field_proxy_poc(m_attr):
    m_attr_readwrite = m_attr + "_readwrite"
    outer = m.Outer()
    field_orig = m.Field()
    field_orig.num = 45
    setattr(outer, m_attr_readwrite, field_orig)
    field_proxy = unique_ptr_field_proxy_poc(outer, m_attr_readwrite)
    assert field_proxy.num == 45
    assert field_proxy.num == 45
    with pytest.raises(AttributeError):
        _ = field_proxy.xyz
    assert field_proxy.num == 45
    field_proxy.num = 82
    assert field_proxy.num == 82
    field_proxy = unique_ptr_field_proxy_poc(outer, m_attr_readwrite)
    assert field_proxy.num == 82
    with pytest.raises(AttributeError):
        del field_proxy.num
    assert field_proxy.num == 82


def test_readonly_char6_member():
    obj = m.WithCharArrayMember()
    assert obj.char6_member == "Char6"


def test_readonly_const_char_ptr_member():
    obj = m.WithConstCharPtrMember()
    assert obj.const_char_ptr_member == "ConstChar*"



# Importing re before pytest after observing a PyPy CI flake when importing pytest first.
from __future__ import annotations

import re

import pytest

from pybind11_tests import class_sh_basic as m


def test_atyp_constructors():
    obj = m.atyp()
    assert obj.__class__.__name__ == "atyp"
    obj = m.atyp("")
    assert obj.__class__.__name__ == "atyp"
    obj = m.atyp("txtm")
    assert obj.__class__.__name__ == "atyp"


@pytest.mark.parametrize(
    ("rtrn_f", "expected"),
    [
        (m.rtrn_valu, "rtrn_valu(_MvCtor)*_MvCtor"),
        (m.rtrn_rref, "rtrn_rref(_MvCtor)*_MvCtor"),
        (m.rtrn_cref, "rtrn_cref(_MvCtor)*_CpCtor"),
        (m.rtrn_mref, "rtrn_mref(_MvCtor)*_CpCtor"),
        (m.rtrn_cptr, "rtrn_cptr"),
        (m.rtrn_mptr, "rtrn_mptr"),
        (m.rtrn_shmp, "rtrn_shmp"),
        (m.rtrn_shcp, "rtrn_shcp"),
        (m.rtrn_uqmp, "rtrn_uqmp"),
        (m.rtrn_uqcp, "rtrn_uqcp"),
        (m.rtrn_udmp, "rtrn_udmp"),
        (m.rtrn_udcp, "rtrn_udcp"),
    ],
)
def test_cast(rtrn_f, expected):
    assert re.match(expected, m.get_mtxt(rtrn_f()))


@pytest.mark.parametrize(
    ("pass_f", "mtxt", "expected"),
    [
        (m.pass_valu, "Valu", "pass_valu:Valu(_MvCtor)*_CpCtor"),
        (m.pass_cref, "Cref", "pass_cref:Cref(_MvCtor)*_MvCtor"),
        (m.pass_mref, "Mref", "pass_mref:Mref(_MvCtor)*_MvCtor"),
        (m.pass_cptr, "Cptr", "pass_cptr:Cptr(_MvCtor)*_MvCtor"),
        (m.pass_mptr, "Mptr", "pass_mptr:Mptr(_MvCtor)*_MvCtor"),
        (m.pass_shmp, "Shmp", "pass_shmp:Shmp(_MvCtor)*_MvCtor"),
        (m.pass_shcp, "Shcp", "pass_shcp:Shcp(_MvCtor)*_MvCtor"),
        (m.pass_uqmp, "Uqmp", "pass_uqmp:Uqmp(_MvCtor)*_MvCtor"),
        (m.pass_uqcp, "Uqcp", "pass_uqcp:Uqcp(_MvCtor)*_MvCtor"),
    ],
)
def test_load_with_mtxt(pass_f, mtxt, expected):
    assert re.match(expected, pass_f(m.atyp(mtxt)))


@pytest.mark.parametrize(
    ("pass_f", "rtrn_f", "expected"),
    [
        (m.pass_udmp, m.rtrn_udmp, "pass_udmp:rtrn_udmp"),
        (m.pass_udcp, m.rtrn_udcp, "pass_udcp:rtrn_udcp"),
    ],
)
def test_load_with_rtrn_f(pass_f, rtrn_f, expected):
    assert pass_f(rtrn_f()) == expected


@pytest.mark.parametrize(
    ("pass_f", "rtrn_f", "regex_expected"),
    [
        (
            m.pass_udmp_del,
            m.rtrn_udmp_del,
            "pass_udmp_del:rtrn_udmp_del,udmp_deleter(_MvCtorTo)*_MvCtorTo",
        ),
        (
            m.pass_udcp_del,
            m.rtrn_udcp_del,
            "pass_udcp_del:rtrn_udcp_del,udcp_deleter(_MvCtorTo)*_MvCtorTo",
        ),
        (
            m.pass_udmp_del_nd,
            m.rtrn_udmp_del_nd,
            "pass_udmp_del_nd:rtrn_udmp_del_nd,udmp_deleter_nd(_MvCtorTo)*_MvCtorTo",
        ),
        (
            m.pass_udcp_del_nd,
            m.rtrn_udcp_del_nd,
            "pass_udcp_del_nd:rtrn_udcp_del_nd,udcp_deleter_nd(_MvCtorTo)*_MvCtorTo",
        ),
    ],
)
def test_deleter_roundtrip(pass_f, rtrn_f, regex_expected):
    assert re.match(regex_expected, pass_f(rtrn_f()))


@pytest.mark.parametrize(
    ("pass_f", "rtrn_f", "expected"),
    [
        (m.pass_uqmp, m.rtrn_uqmp, "pass_uqmp:rtrn_uqmp"),
        (m.pass_uqcp, m.rtrn_uqcp, "pass_uqcp:rtrn_uqcp"),
        (m.pass_udmp, m.rtrn_udmp, "pass_udmp:rtrn_udmp"),
        (m.pass_udcp, m.rtrn_udcp, "pass_udcp:rtrn_udcp"),
    ],
)
def test_pass_unique_ptr_disowns(pass_f, rtrn_f, expected):
    obj = rtrn_f()
    assert pass_f(obj) == expected
    with pytest.raises(ValueError) as exc_info:
        pass_f(obj)
    assert str(exc_info.value) == (
        "Missing value for wrapped C++ type"
        + " `pybind11_tests::class_sh_basic::atyp`:"
        + " Python instance was disowned."
    )


@pytest.mark.parametrize(
    ("pass_f", "rtrn_f"),
    [
        (m.pass_uqmp, m.rtrn_uqmp),
        (m.pass_uqcp, m.rtrn_uqcp),
        (m.pass_udmp, m.rtrn_udmp),
        (m.pass_udcp, m.rtrn_udcp),
    ],
)
def test_cannot_disown_use_count_ne_1(pass_f, rtrn_f):
    obj = rtrn_f()
    stash = m.SharedPtrStash()
    stash.Add(obj)
    with pytest.raises(ValueError) as exc_info:
        pass_f(obj)
    assert str(exc_info.value) == ("Cannot disown use_count != 1 (load_as_unique_ptr).")


def test_unique_ptr_roundtrip():
    # Multiple roundtrips to stress-test instance registration/deregistration.
    num_round_trips = 1000
    recycled = m.atyp("passenger")
    for _ in range(num_round_trips):
        id_orig = id(recycled)
        recycled = m.unique_ptr_roundtrip(recycled)
        assert re.match("passenger(_MvCtor)*_MvCtor", m.get_mtxt(recycled))
        id_rtrn = id(recycled)
        # Ensure the returned object is a different Python instance.
        assert id_rtrn != id_orig
        id_orig = id_rtrn


def test_pass_unique_ptr_cref():
    obj = m.atyp("ctor_arg")
    assert re.match("ctor_arg(_MvCtor)*_MvCtor", m.get_mtxt(obj))
    assert re.match("ctor_arg(_MvCtor)*_MvCtor", m.pass_unique_ptr_cref(obj))
    assert re.match("ctor_arg(_MvCtor)*_MvCtor", m.get_mtxt(obj))


def test_rtrn_unique_ptr_cref():
    obj0 = m.rtrn_unique_ptr_cref("")
    assert m.get_mtxt(obj0) == "static_ctor_arg"
    obj1 = m.rtrn_unique_ptr_cref("passed_mtxt_1")
    assert m.get_mtxt(obj1) == "passed_mtxt_1"
    assert m.get_mtxt(obj0) == "passed_mtxt_1"
    assert obj0 is obj1


def test_unique_ptr_cref_roundtrip():
    # Multiple roundtrips to stress-test implementation.
    num_round_trips = 1000
    orig = m.atyp("passenger")
    mtxt_orig = m.get_mtxt(orig)
    recycled = orig
    for _ in range(num_round_trips):
        recycled = m.unique_ptr_cref_roundtrip(recycled)
        assert recycled is orig
        assert m.get_mtxt(recycled) == mtxt_orig


@pytest.mark.parametrize(
    ("pass_f", "rtrn_f", "moved_out", "moved_in"),
    [
        (m.uconsumer.pass_valu, m.uconsumer.rtrn_valu, True, True),
        (m.uconsumer.pass_rref, m.uconsumer.rtrn_valu, True, True),
        (m.uconsumer.pass_valu, m.uconsumer.rtrn_lref, True, False),
        (m.uconsumer.pass_valu, m.uconsumer.rtrn_cref, True, False),
    ],
)
def test_unique_ptr_consumer_roundtrip(pass_f, rtrn_f, moved_out, moved_in):
    c = m.uconsumer()
    assert not c.valid()
    recycled = m.atyp("passenger")
    mtxt_orig = m.get_mtxt(recycled)
    assert re.match("passenger_(MvCtor){1,2}", mtxt_orig)

    pass_f(c, recycled)
    if moved_out:
        with pytest.raises(ValueError) as excinfo:
            m.get_mtxt(recycled)
        assert "Python instance was disowned" in str(excinfo.value)

    recycled = rtrn_f(c)
    assert c.valid() != moved_in
    assert m.get_mtxt(recycled) == mtxt_orig


def test_py_type_handle_of_atyp():
    obj = m.py_type_handle_of_atyp()
    assert obj.__class__.__name__ == "pybind11_type"


def test_function_signatures(doc):
    assert (
        doc(m.args_shared_ptr)
        == "args_shared_ptr(arg0: m.class_sh_basic.atyp) -> m.class_sh_basic.atyp"
    )
    assert (
        doc(m.args_shared_ptr_const)
        == "args_shared_ptr_const(arg0: m.class_sh_basic.atyp) -> m.class_sh_basic.atyp"
    )
    assert (
        doc(m.args_unique_ptr)
        == "args_unique_ptr(arg0: m.class_sh_basic.atyp) -> m.class_sh_basic.atyp"
    )
    assert (
        doc(m.args_unique_ptr_const)
        == "args_unique_ptr_const(arg0: m.class_sh_basic.atyp) -> m.class_sh_basic.atyp"
    )


def test_unique_ptr_return_value_policy_automatic_reference():
    assert m.get_mtxt(m.rtrn_uq_automatic_reference()) == "rtrn_uq_automatic_reference"


def test_pass_shared_ptr_ptr():
    obj = m.atyp()
    with pytest.raises(RuntimeError) as excinfo:
        m.pass_shared_ptr_ptr(obj)
    assert str(excinfo.value) == (
        "Passing `std::shared_ptr<T> *` from Python to C++ is not supported"
        " (inherently unsafe)."
    )


def test_unusual_op_ref():
    # Merely to test that this still exists and built successfully.
    assert m.CallCastUnusualOpRefConstRef().__class__.__name__ == "LocalUnusualOpRef"
    assert m.CallCastUnusualOpRefMovable().__class__.__name__ == "LocalUnusualOpRef"



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

if TYPE_CHECKING:
    from conftest import SanitizedString

from pybind11_tests import docs_advanced_cast_custom as m


def assert_negate_function(
    input_sequence: Sequence[float],
    target: tuple[float, float],
) -> None:
    output = m.negate(input_sequence)
    assert isinstance(output, tuple)
    assert len(output) == 2
    assert isinstance(output[0], float)
    assert isinstance(output[1], float)
    assert output == target


def test_negate(doc: SanitizedString) -> None:
    assert (
        doc(m.negate)
        == "negate(arg0: collections.abc.Sequence[float]) -> tuple[float, float]"
    )
    assert_negate_function([1.0, -1.0], (-1.0, 1.0))
    assert_negate_function((1.0, -1.0), (-1.0, 1.0))
    assert_negate_function([1, -1], (-1.0, 1.0))
    assert_negate_function((1, -1), (-1.0, 1.0))


def test_docs() -> None:
    ###########################################################################
    # PLEASE UPDATE docs/advanced/cast/custom.rst IF ANY CHANGES ARE MADE HERE.
    ###########################################################################
    point1 = [1.0, -1.0]
    point2 = m.negate(point1)
    assert point2 == (-1.0, 1.0)
